!> @file eiram.f90
!> Main file of the library 
!> @authors Markus Brenneis Markus.Brenneis@uni-duesseldorf.de,
!>          Vladislav Kotov v.kotov@fz-juelich.de (contact person)
!>

!    Copyright (c) 2016  Forschungszentrum Juelich GmbH
!                        Markus Brenneis, Vladislav Kotov 
!
!    This file is part of EIRAM.
!
!    EIRAM is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    EIRAM is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with EIRAM.  If not, see <http://www.gnu.org/licenses/>.
!

!> \mainpage
!>
!> Processing of Atomic and Molecuar data stored in "Janev" format:
!> logarithmic polynomial fits introduced in 
!> <a href="http://www.eirene.de/html/hydhel.html">
!> [R.K. Janev et al. Elementary Processes in Hydrogen-Helium Plasmas (Springer 1987)] </a>. 
!> See Chapter 9 of <a href="http://www.eirene.de/janev_buch.pdf">this document</a>.
!> This is a legacy format which was used in the <a href="http://www.eirene.de"> EIRENE code </a>.
!> E.g. data sets <a href="http://www.eirene.de/hydhel.pdf">HYDHEL</a>, 
!> <a href="http://www.eirene.de/amjuel.pdf">AMJUEL</a>, 
!> <a href="http://www.eirene.de/methane.pdf">METHANE</a>, 
!> <a href="http://www.eirene.de/h2vibr.pdf">H2vibr</a>.  
!>
!>
!> <a href="../eiram_description.pdf">Description of the module</a>.  
!>
!> \page ErrorCodes
!> Convention for error (>0) and warning (<0) codes 
!> -    0  : no errors and warnings  
!> -   50  : module not initialized
!> -   100 : wrong input
!> -   110 : could not find requested reaction
!> -   120 : double fit is requested for reaction described by a single fit
!> -   130 : single fit is requested for reaction described by a double fit 
!> -  -100 : wrong input, corrected  
!> -   200 : memory allocation problem
!> -  -200 : memory deallocation problem
!> -   300 : problem while reading file
!> -   310 : input file is not found
!> -  -300 : problem while writing file
!> -   400 : internal problem
!> -  -400 : ...  
!>
!> \todo  
!>  1. An option to load only selected reactions (like in EIRENE)
!>  2. MPI broadcast
    
!> Elementary function which calculates a single polynomial fit 
!>
!> \f$ y = \sum^n_{i=0} b_i X^i \f$
!>
!> This function is called in 
!> \ref eiram_calc1_fast,
!> \ref eiram_calc1,
!> \ref eiram_calc12,
!> \ref eiram_calc2
!>
!> It was decided to use a function call although it costs additional time because :
!> i) in this case the plynomial calculation code is implemented only once, 
!> ii) compiler can use inlining to optimize the call 
 function eiram_fit(B,N,X) 
    implicit none  
    integer,parameter :: DP=SELECTED_REAL_KIND(13)
    real(DP) :: eiram_fit 
    !> Coefficients of the polynomial 
    real(DP),intent(in) :: B(0:N)
    !> Order of the polynomial 
    integer,intent(in) :: N
    !> Value for which the polynomial is calculated
    real(DP),intent(in) :: X

    integer :: i

    eiram_fit=B(N)
    do i=N-1,0,-1   
     eiram_fit=eiram_fit*X+B(i)
    end do 

 end function eiram_fit

module eiram
    implicit none

    public eiram_load        
    public eiram_deallocate  
    public eiram_get_id       
    public eiram_get_order
    public eiram_create_image
    public eiram_calc1_fast
    public eiram_calc1        
    public eiram_calc12   
    public eiram_calc2   

    !> Kind parameter for real numbers
    integer, parameter, public :: EIRAM_DP = SELECTED_REAL_KIND(13)
    !> Unit to which messages are written
    integer, public :: eiram_unit = 6
    
    !> Length of the string with units
    integer,parameter :: UNITS_LENGTH=32
    !> Length of the string with description of a quantity 
    integer,parameter :: QUANTITY_LENGTH=64
    !> Length of the string with the reaction equation 
    integer,parameter :: REACTION_LENGTH=128
    !> Length of the string with description of the process 
    integer,parameter :: DESCRIPTION_LENGTH=256
    !> Length of the string with a name of an input file
    integer, public, parameter :: EIRAM_FILE_NAME_LENGTH = 32
    !> Length of the string with reactiontype 
    integer, public, parameter :: EIRAM_TYPE_LENGTH = 8
    !> Length of the string with reaction index
    integer, public, parameter :: EIRAM_INDEX_LENGTH = 16
    !> Maximum length of a line read from an file
    integer, parameter :: MAX_LINE_LENGTH = 256

    logical, parameter :: DEBUG = .false.

    external eiram_fit
    real(eiram_dp),public :: eiram_fit
 
    !> Data for one reaction
    !> 
    !> The reaction rate or cross section \f$ Y(X_1,X_2) \f$ is described by the equation:
    !>
    !> \f$ Y(X_1,X_2)=\sum_{i=0}^n \sum^m_{j=0} A_{ji} \ln^i{X_1} \ln^j{X_2} \f$
    !>
    type, public:: eiram_data
        !> Order of polynomial for the 1st index I
        integer :: N
        !> Order of polynomial for the 2nd index J, M=0 for 1D polynomial
        integer :: M
        !> Coefficients of the polynomial, CREAC(I,J)=\f$ A_{ij} \f$
        !>
        !> If M=0 then CREAC(0:N,0) else CREAC(0:M,0:N) 
        !> (that the sums are always calculated over the "fast" 1st index) 
        real(EIRAM_DP), allocatable :: CREAC(:,:)
        !> Minimal value of the 1st variable \f$ X_1 \f$ for which the fit is valid 
        real(EIRAM_DP) :: X1MIN 
        !> Maximal value of the 1st variable \f$ X_1 \f$ for which the fit is valid 
        real(EIRAM_DP) :: X1MAX 
        !> Minimal value of the 2nd variable \f$ X_2 \f$ for which the fit is valid 
        real(EIRAM_DP) :: X2MIN 
        !> Maximal value of the 2nd variable \f$ X_2 \f$ for which the fit is valid 
        real(EIRAM_DP) :: X2MAX
        !> Units of variable \f$ X_1 \f$ in human-readable format
        character(UNITS_LENGTH) :: X1UNITS = ""
        !> Units of variable \f$ X_2 \f$ in human-readable format
        character(UNITS_LENGTH) ::  X2UNITS = "" 
        !> Units of the resulting rate or cross-section \f$ Y \f$ in human-readable format
         character(UNITS_LENGTH) ::  YUNITS = ""
        !> Name of the physical quantity described by \f$ X_1 \f$ in human-readable format
        character(QUANTITY_LENGTH) :: X1QUANTITY = ""
        !> Name of the physical quantity described by \f$ X_2 \f$ in human-readable format
        character(QUANTITY_LENGTH) :: X2QUANTITY = "" 
        !> Name of the physical quantity described by \f$ Y \f$ in human-readable format
        character(QUANTITY_LENGTH) :: YQUANTITY = ""
        !> Equation of the reaction in human-readable format 
        character(REACTION_LENGTH) :: REACTION = ""
        !> Description of the process (human-readable) 
        character(DESCRIPTION_LENGTH) :: DESCRIPTION = ""
        !> Name of the input file (data set): HYDHEL, AMJUEL etc.
        character(EIRAM_FILE_NAME_LENGTH) :: FILE = ""
        !> Type of the reaction: H.1, H.2, H.3 etc.
        character(EIRAM_TYPE_LENGTH) :: TYPE = ""
        !> Index of reaction, e.g. 2.1.5 
        character(EIRAM_INDEX_LENGTH) :: INDEX = ""
    end type eiram_data

    private

    !> current number of loaded reactions
    integer,public :: eiram_nreact=0

    !>Array of reactions
    type(eiram_data), allocatable :: REACTIONS(:)
    
    contains

!***********************************************************************************
!   INITIALIZATION/DEINITIALIZATION ROUTINES  
!***********************************************************************************
    
    !> Initialization of the module from input files (data sets)
    !>
    !> This subroutine reads data from the files HYDHEL, AMJUEL etc. 
    !> The subroutine can be called more than once - 
    !> in this case the data loaded on the previous calls are appended
    !>
    !> Error code 310 is returned if requested input file is not found
    subroutine eiram_load(filePath, fileName, err)
        !> Full path to the input file (w/o its name)
        character(*), intent(in) :: filePath
        !> Name of the input file
        character(*), intent(in) :: fileName
        !> Error code
        integer, intent(out) :: err

        logical :: fileExists,isLaTeX
        !> number of read data sets
        integer :: nData, nOldData
        !> upper bound for number of data sets
        integer :: l,nDataMax
        integer, parameter :: UNIT = 20
        character(MAX_LINE_LENGTH) :: line
        character(*), parameter :: BEGIN_DATA_MARK = "##BEGIN DATA HERE##"
        !string which is used to find out that the document is in LaTeX format
        ! - w/o '\' at the beginning because for some reason this confuses g95 compiler
        character(*), parameter :: LATEX_MARK = "begin{document}"
        type(eiram_data), allocatable:: reactions_(:)

        intrinsic index,trim

         if(DEBUG) print *, "open file ",trim(fileName)

        inquire(file=trim(filePath)//trim(fileName),exist=fileExists,iostat=err)
        if(.not.fileExists .or. err /= 0) then
         err=310
         write(eiram_unit, *) "ERROR in eiram_load: input file ", &
                              trim(filePath)//trim(fileName)," is not found" 
         return
        end if

        if(allocated(REACTIONS)) then
            nOldData = size(REACTIONS)
        else
            nOldData = 0
        end if
        
        nDataMax = countLinesContaining(trim(filePath)//trim(fileName), "Reaction", err)
        if(err/=0) then
            write(eiram_unit, *) "ERROR in eiram_load while counting lines, err ", err
            err = 300
            return
        end if
        
        allocate(reactions_(nDataMax+nOldData),stat=err)
        if(err/=0) then
            write(eiram_unit, *) "ERROR in eiram_load while allocating temporary array, ",&
                                 "nData",nDataMax+nOldData
            err = 200
            return
        end if

        open(UNIT, iostat=err, file=trim(filePath)//trim(fileName), status='old', action='read')
        if(err /= 0) then
            write(eiram_unit, *) "ERROR in eiram_load while reading "//trim(filePath)//trim(fileName)
            close(UNIT)
            err = 300
            return
        end if

        l = 0
        nData = 0
        isLaTeX=.false.        

        !look for beginning of the data section
        do
          read(UNIT, '(A)', iostat=err,end=1000) line
          l = l + 1
          if(err /= 0) goto 1300

          if(index(line,LATEX_MARK) > 0) isLaTeX=.true.

          if(index(line, BEGIN_DATA_MARK) > 0) then
            if(DEBUG) print *, "found BEGIN_DATA_MARK in line ", l
            exit 
          end if         
        end do        

       !read the data 
       if(isLaTeX) then 
        call eiram_read_latex(reactions_,fileName,unit,l,nData,*1300,*1303,*1304)        
       else
        call eiram_read_txt(reactions_,fileName,unit,l,nData,*1300,*1303,*1304)        
       end if
     
       if(DEBUG) print *, "EOF in", l
       close(UNIT)
        
       if(allocated(REACTIONS)) then
          !Add reactions which were already loaded in previous calls of eiram_load
           reactions_(nData+1:nData+nOldData) = REACTIONS(1:nOldData)
           deallocate(REACTIONS, stat=err)
           if(err /= 0) then
             write(eiram_unit, *) "ERROR in eiram_load: cannot re-allocate memory ", err
             err = 200
            return
           end if
       end if
        
        !(Re-)allocate REACTIONS
        allocate(REACTIONS(nData+nOldData),stat=err)
        if(err /= 0) then
         write(eiram_unit, *) "ERROR in eiram_load: cannot allocate the array of reactions ",&
                              "nData ",nData+nOldData
         err = 200
         return
        end if

        REACTIONS(1:nData+nOldData) = reactions_(1:nData+nOldData)
 
        call deallocate_reactions

        write(eiram_unit, *) "initalization from file "//trim(filePath)//trim(fileName)//" completed"
        write(eiram_unit, *) "  new data sets loaded: ", nData
        eiram_nreact=nData+nOldData
        write(eiram_unit, *) "  total number of data sets: ", eiram_nreact
        return

1000    write(eiram_unit, *) "ERROR in eiram_load:  no data section is found in file "//trim(filePath)//trim(fileName)
        close(UNIT)
        call deallocate_reactions
        err = 300
        return
1300    write(eiram_unit, *) "ERROR in eiram_load:  while reading line ", l, " of "//trim(filePath)//trim(fileName)
        close(UNIT)
        call deallocate_reactions
        err = 300
        return
1303    write(eiram_unit, *) "ERROR in eiram_load: wrong format in line ", l, & 
                             " of "//trim(filePath)//trim(fileName), err
        close(UNIT)
        call deallocate_reactions
        err = 300
        return
1304    write(eiram_unit, *) "ERROR in eiram_load: cannot allocate CREAC "
        write(eiram_unit, *) " line ", l, " of ",trim(filePath)//trim(fileName),&
                             " reaction ",trim(reactions_(nData)%TYPE)//" "//trim(reactions_(nData)%INDEX)
        close(UNIT)
        call deallocate_reactions
        err = 200
        return       

    contains 

    !clean-up for the case if the compiler uses argument -save       
     subroutine deallocate_reactions
      integer :: i
      logical :: lerr
      lerr=.false.
      do i=1,size(reactions_)
       if(allocated(reactions_(i)%CREAC)) then
         deallocate(reactions_(i)%CREAC,stat=err)
         if(err /= 0) lerr=.true.
       end if  
      end do   
      deallocate(reactions_,stat=err)
      if(err /= 0 .or. lerr ) then
        write(eiram_unit, *) "WARNING from eiram_load: problem while deallocating temporary array "
        err = -200          
      end if
     end subroutine deallocate_reactions

    end subroutine eiram_load

    !>
    !> Read data from a LaTeX file
    !>
    subroutine eiram_read_latex(reactions_,fileName,unit,l,nData,*,*,*)
     type(eiram_data), allocatable:: reactions_(:)
     character(*), intent(in) :: fileName
     integer,intent(in) :: unit  
     integer,intent(inout) :: l 
     integer,intent(out) :: nData    
     integer :: gotoLine,numberOfUnclosedBraces,n,err,i1,i2
     character(EIRAM_TYPE_LENGTH) :: stype
     character(MAX_LINE_LENGTH) :: line
     character(MAX_LINE_LENGTH*2) :: reactionName

     intrinsic index,count,transfer,trim,adjustl,len_trim

       nData=0

       forEachLine: do
       read(UNIT, '(A)', iostat=err,end=100) line
       l = l + 1
       if(err /= 0) return 1
       newSection: if(index(line, "\section") == 1) then
         if(DEBUG) print *, "found \section in line ", l
         if(index(trim(line), " ") == 0) then
           read(UNIT, '(A)', iostat=err) line
           l = l + 1
           if(err /= 0) return 1
          end if
          stype= line(index(line,"{")+1:index(line," ")-1)
          if(index(stype,":")>0) stype=stype(1:index(stype,":")-1)
       end if newSection

       newReaction: if(index(line, "subsection{") > 0) then
            numberOfUnclosedBraces = count(transfer(line, 'a', MAX_LINE_LENGTH) == "{")
            n = count(transfer(line, 'a', MAX_LINE_LENGTH) == "}")
            numberOfUnclosedBraces = numberOfUnclosedBraces - n
            reactionName = trim(adjustl(line))
 
            !for the case when the reaction is defined like this 
            !\subsection{
            !             ............ }              
            findEndOfReactionName: do
                if(numberOfUnclosedBraces == 0) exit findEndOfReactionName
                read(UNIT, '(A)', iostat=err) line
                l = l + 1
                if(err /= 0) return 1
                n = count(transfer(line, 'a', MAX_LINE_LENGTH) == "{")
                numberOfUnclosedBraces = numberOfUnclosedBraces + n
                n = count(transfer(line, 'a', MAX_LINE_LENGTH) == "}")
                numberOfUnclosedBraces = numberOfUnclosedBraces - n
                if(len_trim(reactionName) /= index(reactionName, "{")) then
                 reactionName = trim(reactionName)//" "//trim(adjustl(line))
                else
                 reactionName = trim(reactionName)//trim(adjustl(line))
                end if
                        if(line == "}") exit findEndOfReactionName
             end do findEndOfReactionName
                          
             if(index(reactionName,'Reaction')<1) cycle

             nData = nData+1

             reactions_(nData)%FILE = fileName 
             reactions_(nData)%TYPE = stype

             if(DEBUG) print *, "found Reaction in line ", l
             !Sub-string between 1st and 2nd space after "Reaction": the index 
             i1=index(reactionName,"Reaction")+8;
             reactions_(nData)%INDEX = reactionName(index(reactionName(i1:)," ")+i1:)
             reactions_(nData)%INDEX = reactions_(nData)%INDEX(:index(reactions_(nData)%INDEX," ")-1)
             if(DEBUG) print *, "  with index ", trim(reactions_(nData)%INDEX)
             !Sub-string between $ and $ space: reaction equation (if available)
             i1=index(reactionName,"$")
             i2=index(reactionName,"$",.true.)
             if (i1>0 .and. i2>0 ) then
               reactions_(nData)%REACTION = reactionName(i1:i2)
               if(DEBUG) print *, "  with equation ", trim(reactions_(nData)%REACTION)
             end if 
                   
             skipTillVerbatim: do
                read(UNIT, '(A)', iostat=err) line
                l = l + 1
                if(err /= 0) return 1
                if(index(line, "verbatim") > 0) then
                   exit skipTillVerbatim
                end if
             end do skipTillVerbatim
                                       
             select case(reactions_(nData)%TYPE)
                    case("H.1", "H.2", "H.8", "H.11")
                         if(DEBUG) print *, "h1/2/8/11"                
                         call parse_single(reactions_(nData),unit,gotoLine,l)
                         if(gotoLine == 1300) return 1
                         if(gotoLine == 1303) return 2
                         if(gotoLine == 1304) return 3
                         if(.not.allocated(reactions_(nData)%CREAC)) nData=nData-1 !skip this reaction
                    case ("H.3", "H.4", "H.9", "H.10", "H.12")
                         if(DEBUG) print *, "h3/4/9/10/12"
                         call parse_double(reactions_(nData),unit,gotoLine,l)
                         if(gotoLine == 1300) return 1
                         if(gotoLine == 1303) return 2
                         if(gotoLine == 1304) return 3
                         if(.not.allocated(reactions_(nData)%CREAC)) nData=nData-1 !skip this reaction
                    case default
       !                  write(eiram_unit, *) "WARNING from eiram_load: No handle for type "&
       !                                       //trim(reactions_(nData)%TYPE)//" in line", l
                          if(DEBUG) print *, ">> exiting"
                          nData = nData-1 !skip reaction
             end select                    
         end if newReaction
            
        end do forEachLine

  100   continue

    end subroutine eiram_read_latex

    !>
    !> Read data from a plain text file
    !>
    subroutine eiram_read_txt(reactions_,fileName,unit,l,nData,*,*,*)
     type(eiram_data), allocatable:: reactions_(:)
     character(*), intent(in) :: fileName
     integer,intent(in) :: unit  
     integer,intent(inout) :: l 
     integer,intent(out) :: nData    
     integer :: gotoLine,err,i1,i2,itype
     character(EIRAM_TYPE_LENGTH) :: stype
     character(MAX_LINE_LENGTH) :: line, reactionName

     intrinsic index,trim,adjustl,len_trim

     nData=0
     itype=0

     forEachLine: do
       read(UNIT, '(A)', iostat=err,end=100) line
       l = l + 1
       if(err /= 0) return 1
       call gettype(line,stype)

      newReaction: if(index(line, "Reaction") > 0) then
         nData = nData+1
         reactions_(nData)%FILE = fileName 
         reactions_(nData)%TYPE = stype
         reactionName = trim(adjustl(line))
         !Sub-string between 1st and 2nd space: the index 
         i1=index(reactionName," ")+1 
         i2=index(reactionName(i1:)," ")+i1-2
         reactions_(nData)%INDEX = reactionName(i1:i2)          
         !Sub-string between 2nd space and end of the line: reaction equation
         reactions_(nData)%REACTION =trim(adjustl(reactionName(i2+1:)))
         if(DEBUG) then 
          print *, "found new reaction"
          print *, "             index ",trim(reactions_(nData)%INDEX)
          print *, "             equation ",trim(reactions_(nData)%REACTION)
         end if
         select case(reactions_(nData)%TYPE)
                case("H.1", "H.2", "H.8", "H.11")
                     if(DEBUG) print *, "h1/2/8/11"                
                     call parse_single(reactions_(nData),unit,gotoLine,l)
                     if(gotoLine == 1300) return 1
                     if(gotoLine == 1303) return 2
                     if(gotoLine == 1304) return 3
                     if(.not.allocated(reactions_(nData)%CREAC)) nData=nData-1 !skip this reaction
                case ("H.3", "H.4", "H.9", "H.10", "H.12")
                     if(DEBUG) print *, "h3/4/9/10/12"
                     call parse_double(reactions_(nData),unit,gotoLine,l)
                     if(gotoLine == 1300) return 1
                     if(gotoLine == 1303) return 2
                     if(gotoLine == 1304) return 3
                     if(.not.allocated(reactions_(nData)%CREAC)) nData=nData-1 !skip this reaction
                case default
       !              write(eiram_unit, *) "WARNING from eiram_load: No handle for type "&
       !                                   //trim(reactions_(nData)%TYPE)//" in line", l
                if(DEBUG) print *, ">> exiting"
                nData = nData-1 !skip reaction
         end select     
      end if newReaction
     end do forEachLine

100  continue

    contains 

    !> Set the reaction type if new section is reached
    subroutine gettype(line,stype)
     character(MAX_LINE_LENGTH),intent(in) :: line
     character(EIRAM_TYPE_LENGTH),intent(inout) :: stype

     integer,parameter :: ntp=12
     integer :: i 
     character(4),parameter :: stp(0:ntp)= &
         (/'H.0 ','H.1 ','H.2 ','H.3 ','H.4 ','H.5 ','H.6 ','H.7 ','H.8 ','H.9 ','H.10','H.11','H.12'/)
          
     if(index(line,'H.')<1) return !for speed up

     do i=ntp,itype,-1
      if(index(line,trim(stp(i)))>0) then
       stype=trim(stp(i))
       itype=i
       if(DEBUG) print *, "found new type ", stype
       return 
      end if
     end do
 
    end subroutine gettype

    end subroutine eiram_read_txt

    
    !> Deallocate dynamic arrays used by this module
    subroutine eiram_deallocate(err)
        !> Error code 
        integer, intent(out) :: err
        
        logical :: lerr 
        integer i       

        intrinsic size

        err = 0

        eiram_nreact=0

        if(allocated(REACTIONS)) then 

         lerr=.false.
         do i=1,size(REACTIONS)
          if(allocated(REACTIONS(i)%CREAC)) then
           deallocate(REACTIONS(i)%CREAC,stat=err)
           if(err /= 0) lerr=.true.
          end if  
         end do            

         deallocate(REACTIONS, stat=err)
         if(err /= 0) then
            write(eiram_unit, *) "WARNING from eiram_deallocate: problem while deallocating memory, err ", err
            err = -200
            return
         end if

         if(lerr) then
          write(eiram_unit, *) "WARNING from eiram_deallocate: problem while deallocating memory"
          err=-200
         end if   

       end if !if(allocated(allocated(REACTIONS)) then

    end subroutine eiram_deallocate

    !> Parse input for single parameter reactions       
    subroutine parse_single(reaction,unit,gotoLine,l)
     type(eiram_data) :: reaction
     integer,intent(in) :: unit 
     integer,intent(out) :: gotoLine
     integer,intent(inout) :: l  

     character(2) :: chr
     character(80) :: zeile
     integer :: err,indff,iftflg,ind,ind2,i,j
    
     intrinsic index,mod

     gotoLine = 0     
             
     reaction%M = 0
     reaction%N = 8
     if(reaction%TYPE == "H.1") then
        reaction%YUNITS = "cm^2"
        reaction%YQUANTITY = "cross section"
        reaction%X1QUANTITY = "Energy"
        chr='a0'
     elseif(reaction%TYPE == "H.2") then
        reaction%YUNITS = "cm^3/s"
        reaction%YQUANTITY = "collision rate"
        reaction%X1QUANTITY = "Temperature"
        chr='b0'
     elseif(reaction%TYPE == "H.8") then
        reaction%YUNITS = "cm^3/s\cdot eV"
        reaction%YQUANTITY = "electron cooling rate"
        reaction%X1QUANTITY = "Temperature"
        chr='h0'
     elseif(reaction%TYPE == "H.11") then
        reaction%YUNITS = ""
        reaction%YQUANTITY = "ratio"
        reaction%X1QUANTITY = "Temperature"
        chr='k0'
     end if
     reaction%X1UNITS = "eV"

     !this code mimics EIRENE, slreac.f 
3    read(unit,'(A80)',iostat=err) zeile
     l=l+1
     if(err /= 0) then
       gotoLine = 1300
       return
     end if

     iftflg=0

     !Skip "An analytic formula is given in the text" in LaTeX mode
     if( index(zeile,'verbatim')>0 .or. index(zeile,'subsection')>0 ) return

     indff=index(zeile,'fit-flag')
     if (index(zeile,chr)+indff.EQ.0) GOTO 3
     if(indff > 0) THEN
       read (zeile((indff+8):80),*,iostat=err) iftflg
       if(err /= 0) then
          gotoLine = 1303
          return
       end if
       goto 3
      end if

      allocate(reaction%CREAC(0:8,0:0),stat=err)
      if(err /=0) then
        gotoLine=1304
        return
      end if

      if(mod(iftflg,100) == 10) then
        ind=index(zeile,chr(1:1))
        reaction%creac(:,0)=0.
        read(zeile((ind+2):80),*,iostat=err) reaction%creac(0,0)
        if(err /= 0) then
          gotoLine = 1303
          return
        end if
      else
        do j=0,2
         ind=index(zeile,chr(1:1))+2                   
         do i=0,2
          ind2=index(zeile(ind:80),chr(1:1))
          if(ind2<1) then
           ind2=80
          else
           ind2=ind2+ind-2
          end if 
          read(zeile(ind:ind2),*,iostat=err) reaction%creac(j*3+i,0)          
          if(err /= 0) then
           gotoLine = 1303
           return
          end if 
          ind=ind2+3
         end do
         read(unit,'(A80)',iostat=err) zeile
         l=l+1
         if(err /= 0) then
          gotoLine = 1300
          return
         end if
        end do
      end if
               
      if(DEBUG) print *, reaction%CREAC         

     end subroutine parse_single

     !>
     !> Parse input for two parameter reactions       
     !>
     subroutine parse_double(reaction,unit,gotoLine,l)
      type(eiram_data) :: reaction
      integer,intent(in) :: unit 
      integer,intent(out) :: gotoLine
      integer,intent(inout) :: l  
 
      character(80) :: zeile
      integer :: err,indff,iftflg,ih,i,j,k

      intrinsic index,mod

      gotoLine = 0
            
      reaction%N = 8
      reaction%M = 8
      if(reaction%TYPE == "H.3") then
         reaction%YQUANTITY = "collision rate"
         reaction%YUNITS = "cm^3/s"
         reaction%X1QUANTITY = "Energy"
         reaction%X1UNITS = "eV"
         reaction%X2QUANTITY = "Temperature"
         reaction%X2UNITS = "eV"
         reaction%X2QUANTITY = "Temperature"
      elseif(reaction%TYPE == "H.4") then
         reaction%YQUANTITY = "collision rate"
         reaction%YUNITS = "cm^3/s"
         reaction%X1QUANTITY = "Density"
         reaction%X1UNITS = "cm^{-3}"
         reaction%X2QUANTITY = "Temperature"
         reaction%X2UNITS = "eV"
      elseif(reaction%TYPE == "H.9") then
         reaction%YQUANTITY = "ion cooling rate"
         reaction%YUNITS = "cm^3/s\cdot eV"
         reaction%X1QUANTITY = "Energy"
         reaction%X1UNITS = "[eV]"
         reaction%X2QUANTITY = "Temperature"
         reaction%X2UNITS = "eV"
      elseif(reaction%TYPE == "H.10") then
         reaction%YQUANTITY = "electron cooling rate"
         reaction%YUNITS = "cm^3/s\cdot eV"
         reaction%X1QUANTITY = "Density"
         reaction%X1UNITS = "cm^{-3}"
         reaction%X2QUANTITY = "Temperature"
         reaction%X2UNITS = "eV" 
      elseif(reaction%TYPE == "H.12") then
         reaction%YQUANTITY = "ratio"
         reaction%YUNITS = ""
         reaction%X1QUANTITY = "Density"
         reaction%X1UNITS = "cm^{-3}"
         reaction%X2QUANTITY = "Temperature"
         reaction%X2UNITS = "eV" 
      end if
            
      allocate(reaction%CREAC(0:8,0:8),stat=err)
      if(err /=0) then
         gotoLine=1304
         return
      end if

      iftflg=0

      !this code mimics EIRENE, slreac.f 
      do j=0,2
16      read(unit,'(A80)',iostat=err) zeile
        l=l+1
        if(err /= 0) then
           gotoLine = 1300
           return
        end if

       !Skip "An analytic formula is given in the text" in LaTeX mode
        if( index(zeile,'verbatim')>0 .or. index(zeile,'subsection')>0 ) return

        indff=index(zeile,'fit-flag')
        if(index(zeile,'Index')+indff==0) goto 16
        if(indff > 0) then
          read(zeile((indff+8):80),*,iostat=err) iftflg
          if(err /= 0) then
             gotoLine = 1303
             return
           end if 
          goto 16
        end if

        read(unit,'(1X)',iostat=err)
        l=l+1
        if(err /= 0) then
           gotoLine = 1300
           return
        end if

        if(mod(iftflg,100) == 10) then
          read(unit,*,iostat=err) ih,reaction%CREAC(0,0)
          l=l+1
          if(err /= 0) then
            gotoLine = 1300
            return
          end if
          exit
        else
         do i=0,8
          read(unit,*,iostat=err) ih,(reaction%CREAC(i,k),k=j*3,j*3+2)
          if(err /= 0) then
            gotoLine = 1300
            return
          end if
         end do
        end if
      end do
           
      if(DEBUG) print *, reaction%CREAC

      end subroutine parse_double


!***********************************************************************************
!   GET INFORMATION   
!***********************************************************************************
    
    !> Return the ID-index of the given reaction in the arrays of the module 
    !>
    !> This Id must be used then in \ref eiram_calc  
    integer function eiram_get_id(fileName, reactionType, reactionIndex, err)
        !> Name of the input file (dat set): HYDHEL, AMJUEl etc.
        character(*), intent(in) :: fileName
        !> Type of the reaction: H.1, H.2 etc.
        character(*), intent(in) :: reactionType
        !> Index of the reaction: 2.1.5 etc.
        character(*), intent(in) :: reactionIndex
        !> Error code 
        integer, intent(out) :: err
 
        integer :: i       

        eiram_get_id = 0
        if(.not.allocated(REACTIONS)) then
            write(eiram_unit, *) "ERROR in eiram_get_id: EIRAM is not initialized"
            err = 50
            return
        end if
        
        err = 0
        
        do i = 1, size(REACTIONS)
            if(REACTIONS(i)%FILE == fileName &
              .and. REACTIONS(i)%TYPE == reactionType &
              .and. REACTIONS(i)%INDEX == reactionIndex) then
                eiram_get_id=i
                return
            end if
        end do
        
        eiram_get_id = 0
        write(eiram_unit, *) "ERROR in eiram_get_id: reaction "//&
            trim(fileName)//" "//trim(reactionType)//" "//trim(reactionIndex)//&
            " not found"
        err = 110
    end function eiram_get_id

    !> Return order of the polynomial for both variables
    subroutine eiram_get_order(N,M,Id,err)
     !> Order of polynomial for X1
     integer,intent(out) :: N
     !> Order of polynomial for X2
     integer,intent(out) :: M
     !> Index of the reaction (returned by \ref eiram_getId)
     integer,intent(in) :: Id 
     !> Error code 
     integer,intent(out) :: err

     intrinsic size

     err=0

     if(.not.allocated(REACTIONS)) then
         write(eiram_unit, *) "ERROR in eiram_get_order: EIRAM is not initialized"
         err = 50
         return
     end if

     if(id<0 .or. id>size(REACTIONS) ) then 
       write(eiram_unit,*) "ERROR in eiram_get_order: reaction index is out of range"
       write(eiram_unit,*) " index, min, max ",Id,1,size(REACTIONS)
       err = 100
       return
     end if 

     N=REACTIONS(Id)%N
     M=REACTIONS(Id)%M

    end subroutine eiram_get_order

    !> Create a copy (image) of the reactions array stored in the module 
    !>
    !> WARNING: array data which is already allocated will be overwritten.  
    !> In case of error (err /= 0) not allocated array is returned
    subroutine eiram_create_image(data,err)
     !> Resulting image of the module
     type(eiram_data), allocatable :: data(:)
     !> Error code
     integer,intent(out) :: err    

     integer :: k,L,M,N,err0

     intrinsic size,trim 

     err = 0

     if(.not.allocated(REACTIONS)) then
        write(eiram_unit, *) "ERROR in eiram_create_image: EIRAM is not initialized"
        err = 50
        return
     end if

     L=size(REACTIONS)   

     if(allocated(data)) then 
      do k=1,L
       if(allocated(data(k)%CREAC)) then
        deallocate(data(k)%CREAC,stat=err)
        if(err /=0) then 
         write(eiram_unit, *) "ERROR in eiram_create_image: cannot deallocate memory"
         err=200
         return
        end if  
       end if
      end do
      deallocate(data,stat=err)
      if(err /=0) then
       write(eiram_unit, *) "ERROR in eiram_create_image: cannot deallocate memory"
       err=200
       return
      end if
     end if 
     
     allocate(data(L),stat=err)
     if(err /= 0) then
       write(eiram_unit, *) "ERROR in eiram_create_image: cannot allocate memory for the array of reactions"
       write(eiram_unit, *) " Total number of reactions = ",L
       err = 200
       return
     end if

     do k=1,L
      if(.not.allocated(REACTIONS(k)%CREAC)) goto 300
      N=REACTIONS(k)%N
      M=REACTIONS(k)%M
      data(k)%N=N
      data(k)%M=M      
      if(M>0) then
       allocate(data(k)%creac(0:M,0:N),stat=err)
      else 
       allocate(data(k)%creac(0:N,0:0),stat=err)
      end if
      if(err /= 0) goto 200
      data(k)%creac=REACTIONS(k)%CREAC
      data(k)%X1MIN = REACTIONS(k)%X1MIN
      data(k)%X1MAX = REACTIONS(k)%X1MAX
      data(k)%X2MIN = REACTIONS(k)%X2MIN
      data(k)%X2MAX = REACTIONS(k)%X2MAX
      data(k)%X1UNITS =REACTIONS(k)%X1UNITS 
      data(k)%X2UNITS =REACTIONS(k)%X2UNITS
      data(k)%YUNITS = REACTIONS(k)%YUNITS 
      data(k)%X1QUANTITY =REACTIONS(k)%X1QUANTITY
      data(k)%X2QUANTITY =REACTIONS(k)%X2QUANTITY
      data(k)%YQUANTITY = REACTIONS(k)%YQUANTITY
      data(k)%REACTION =REACTIONS(k)%REACTION
      data(k)%DESCRIPTION =REACTIONS(k)%DESCRIPTION
      data(k)%FILE =REACTIONS(k)%FILE 
      data(k)%TYPE =REACTIONS(k)%TYPE 
      data(k)%INDEX =REACTIONS(k)%INDEX 
     end do

     return

200  continue
     write(eiram_unit, *) "ERROR in eiram_create_image: error wile allocating memory"
     write(eiram_unit, *) " Reaction ID, File, Type, Index ",&
                            k,REACTIONS(k)%FILE, REACTIONS(k)%TYPE, REACTIONS(k)%INDEX 
     write(eiram_unit, *) " N, M ",N,M       
     err = 200
     do k=1,L
      if(allocated(data(k)%CREAC)) deallocate(data(k)%CREAC,stat=err0)
     end do
     deallocate(data,stat=err0)
     return
300  continue
     write(eiram_unit, *) "ERROR detected in eiram_create_image: ", &
                          "array of coefficients is not allocated"
     write(eiram_unit, *) " Reaction ID, File, Type, Index ",&
                            k,REACTIONS(k)%FILE,REACTIONS(k)%TYPE,REACTIONS(k)%INDEX 
     err = 50
     do k=1,L
      if(allocated(data(k)%CREAC)) deallocate(data(k)%CREAC,stat=err0)
     end do
     deallocate(data,stat=err0)
     return
    end subroutine eiram_create_image
     
!***********************************************************************************
!   CALCULATE FITS 
!***********************************************************************************

 !> Calculate single polynomial fit: scalar version W/O CHECKS
 !> 
 !> This function is intended for calculation of cross-sections
 !> during a Monte-Carlo particle tracing
    real(EIRAM_DP) function eiram_calc1_fast(Id,lnX)
     !> ID of the reaction returned by eiram_getId
     integer, intent(in) :: Id
     !> Natural logarithm of the argument 
     real(EIRAM_DP), intent(in) :: lnX

     intrinsic exp

     eiram_calc1_fast = eiram_fit( REACTIONS(Id)%CREAC(:,0),REACTIONS(Id)%N,lnX)
     eiram_calc1_fast = exp(eiram_calc1_fast)

     end function eiram_calc1_fast

 !> Calculate single polynomial fit
 !>
 !> Opposite to \ref eiram_calc1_fast this subroutine performs checks and 
 !> returns an array. It is indended for plotting and data analysis 
 !> rather than for execution during a Monte-Carlo particle tracing
    subroutine eiram_calc1(Y,Id,lnX,err)
        !> Array of the results: Y(k) for each LnX(k)     
        real(EIRAM_DP), intent(out) :: Y(:) 
        !> ID of the reaction returned by eiram_getId
        integer, intent(in) :: Id
        !> Natural logarithm of the argument 
        real(EIRAM_DP), intent(in) :: lnX(:)
        !> Error code
        integer, intent(out) :: err

        integer :: k,N,M,L
 
        intrinsic size,exp

        err = 0

        L=size(Y)
        if(L /= size(LnX)) then
          write(eiram_unit,*) "ERROR in eiram_calc1: arrays have different lengths"
          write(eiram_unit,*) " length(Y), length(X) ",size(Y),size(lnX)
          err = 100
          return
        end if

        if(.not.allocated(REACTIONS)) then
            write(eiram_unit, *) "ERROR in eiram_calc1: EIRAM is not initialized"
            err = 50
            return
        end if

        M=size(REACTIONS)
        if(id<0 .or. id>M ) then 
          write(eiram_unit,*) "ERROR in eiram_calc1: reaction index is out of range"
          write(eiram_unit,*) " index, min, max ",Id,1,M
          err = 100
          return
        end if 

        if(.not.allocated(REACTIONS(Id)%CREAC)) then
          write(eiram_unit,*) "ERROR in eiram_calc1: array of fit coefficients is not allocated, Id ",Id
          err = 50
          return
        end if

        M=REACTIONS(Id)%M

        if(M>0) then
          write(eiram_unit,*) "ERROR in eiram_calc1: wrong type of reaction"
          write(eiram_unit,*) " Single polynomial fit is requested for reaction ",&
                              " described by a double polynomial fit" 
          err = 130
          return    
        end if 

        N=REACTIONS(Id)%N    
        do k=1,L   
         Y(k)=eiram_fit( REACTIONS(Id)%CREAC(:,0),N,lnX(k))
         Y(k)=exp(Y(k))
        end do

    end subroutine eiram_calc1

 !> Reduce double polynomial fit to single polynomials 
 !> for given values of the second variable 
 !>
 !> \f$ b_i = \sum^m_{j=0} a_{ji}\log^j{X_2} \f$
 !>
 !> Then:
 !>  
 !> \f$ \log{Y} = \sum^n_{i=0} b_i\log^i{X_1} \f$
 !>
 !> 2nd dimension of B must be equal to the length of lnX,
 !> 1st dimension of B must be equal to the order of 
 !> polynomial N plus 1. N is returned by \ref eiram_get_order
 !> 
 !> This subroutine can be used e.g. in Monte-Carlo sampling
 !> for H.3 reactions. The background temperature is fixed
 !> and does not change. Coefficients B are calculated in advance 
 !> for the temperatures in each cell. During the sampling the 
 !> collision rate is calculated using the equation for \f$ \log{Y} \f$
 !> above: with \f$ X_1 \f$ =  energy of the test particle. 
 !> \ref eiram_fit can be used for that purpose 
    subroutine eiram_calc12(B,Id,lnX,err)
    !> Coefficients of the reduced fit \f$b_i\f$
    !>
    !> 1st index is \f$i\f$=0..N, 2nd index is k: lnX(k)
     real(EIRAM_DP), intent(out) :: B(:,:) 
    !> ID of the reaction returned by eiram_getId
     integer, intent(in) :: Id
    !> Natural logarithm of the 2nd argument X2
     real(EIRAM_DP), intent(in) :: lnX(:)
    !> Error code
     integer, intent(out) :: err

     integer i,k,L,M,N

     intrinsic size

     err=0  

     L=size(B,2)
     if(L /= size(LnX)) then
       write(eiram_unit,*) "ERROR in eiram_calc12: arrays have different lengths"
       write(eiram_unit,*) " length(B), length(X) ",L,size(lnX)
       err = 100
       return
     end if

     if(.not.allocated(REACTIONS)) then
         write(eiram_unit, *) "ERROR in eiram_calc12: EIRAM is not initialized"
         err = 50
         return
     end if

     M=size(REACTIONS)
     if(id<0 .or. id>M ) then 
       write(eiram_unit,*) "ERROR in eiram_calc12: reaction index is out of range"
       write(eiram_unit,*) " index, min, max ",Id,1,M
        err = 100
        return
     end if 

     if(.not.allocated(REACTIONS(Id)%CREAC)) then
      write(eiram_unit,*) "ERROR in eiram_calc12: array of fit coefficients is not allocated, Id ",Id
      err = 50
      return
     end if

     M=REACTIONS(Id)%M
     if(M<1) then
       write(eiram_unit,*) "ERROR in eiram_calc12: wrong type of reaction"
       write(eiram_unit,*) " Double polynomial fit is requested for reaction ",&
                           " described by a single polynomial fit" 
       err = 120
      return    
     end if 
              
     N=REACTIONS(Id)%N
     if( N+1 /= size(B,1) ) then
       write(eiram_unit,*) "ERROR in eiram_calc12: mismatch of array sizes"
       write(eiram_unit,*) " order of polynomial  ",N," requested order ",size(B,1)-1
       err = 100
       return
     end if

     do k=1,L 
      do i=0,N
        B(i+1,k)=eiram_fit(REACTIONS(Id)%CREAC(:,i),M,LnX(k))
      end do 
     end do 

    end subroutine eiram_calc12

 !> Calculate double polynomial fit    
 !>
 !> Lengths of Y and LnX2 must be equal.
 !> Length of LnX1 must be either equal to the length of Y,
 !> or equal to 1. 
 !> In this latter case all Y are calculated for the same value of LnX1
    subroutine eiram_calc2(Y,Id,LnX1,LnX2,err)
        !> Array of the results: Y(k) for each pair (LnX1(k),LnX2(k))     
        real(EIRAM_DP), intent(out) :: Y(:) 
        !> ID of the reaction returned by eiram_getId
        integer, intent(in) :: Id
        !> Natural logarithm of the 1st argument 
        real(EIRAM_DP), intent(in) :: lnX1(:)
        !> Natural logarithm of the 2nd argument 
        real(EIRAM_DP), intent(in) :: lnX2(:)
        !> Error code
        integer, intent(out) :: err

        integer :: k,i,N,M,L,q,qk
       
        intrinsic size,exp,min

        err = 0

        L=size(Y)
        if(L /= size(LnX2)) then
          write(eiram_unit,*) "ERROR in eiram_calc2: arrays have different lengths"
          write(eiram_unit,*) " length(Y), length(X2) ",size(Y),size(lnX2)
          err = 100
          return
        end if
 
        q=size(LnX1) 
        if(q /=L .and. q /=1) then
          write(eiram_unit,*) "ERROR in eiram_calc2: arrays have wrong length"
          write(eiram_unit,*) " length(X1), length(Y),  ",size(lnX1),size(Y)
          write(eiram_unit,*) " Length of X1 this array must be either = length(Y), or = 1"
          err = 100
          return
        end if 

        if(.not.allocated(REACTIONS)) then
            write(eiram_unit, *) "ERROR in eiram_calc2: EIRAM is not initialized"
            err = 50
            return
        end if

        M=size(REACTIONS)
        if(id<0 .or. id>M) then 
          write(eiram_unit,*) "ERROR in eiram_calc2: reaction index is out of range"
          write(eiram_unit,*) " index, min, max ",Id,1,M
          err = 100
          return
        end if 

        if(.not.allocated(REACTIONS(Id)%CREAC)) then
         write(eiram_unit,*) "ERROR in eiram_calc 2: array of fit coefficients is not allocated, Id ",Id
         err = 50
         return
        end if

        M=REACTIONS(Id)%M
        if(M<1) then
          write(eiram_unit,*) "ERROR in eiram_calc2: wrong type of reaction"
          write(eiram_unit,*) " Double polynomial fit is requested for reaction ",&
                              " described by a single polynomial fit" 
          err = 120
          return    
        end if 
            
       N=REACTIONS(Id)%N

       do k=1,L
        Y(k)=eiram_fit(REACTIONS(Id)%CREAC(:,N),M,lnX2(k))
        qk=min(q,k)
        do i=N-1,0,-1
         Y(k)= Y(k)*lnX1(qk)+eiram_fit(REACTIONS(Id)%CREAC(:,i),M,lnX2(k))
        end do
        Y(k)=exp(Y(k))            
       end do      

    end subroutine eiram_calc2

!***********************************************************************************
!   AUXILIARY
!***********************************************************************************

    !> Return the number of lines in the file which contain certain sub-string
    integer function countLinesContaining(filePath, string, err)
        !> File to be scanned
        character(*), intent(in) :: filePath
        !> (Sub-)string being looked for 
        character(*), intent(in) :: string
        !> Error code
        integer, intent(out) :: err

        integer, parameter :: UNIT = 22
        character(MAX_LINE_LENGTH) :: line

        intrinsic trim

        open(UNIT, iostat=err,  file=filePath, status='old', action='read')
        countLinesContaining = 0
        if(err /= 0) then
            write(eiram_unit, *) "ERROR in countLinesContaining while reading "//trim(filePath), err
            close(UNIT)
            err = 300
            return
        end if
        do
          read(UNIT, '(A)', iostat=err,end=100) line
          if(err /= 0) then
             write(eiram_unit, *) "ERROR in countLinesContaining while reading line of "//trim(filePath)
            close(UNIT)
             err = 300
             return
          end if
          if(index(line, string) > 0) countLinesContaining = countLinesContaining + 1
        end do 
100     err = 0
        close(UNIT)
    end function countLinesContaining
    
end module eiram

