!> @file eiram_interfaces.f90
!> Wrappers which are used to call \ref eiram from C/C++, as well as in Matlab and Octave interfaces
!> @authors Markus Brenneis Markus.Brenneis@uni-duesseldorf.de,
!>          Vladislav Kotov v.kotov@fz-juelich.de (contact person)

!    Copyright (c) 2016  Forschungszentrum Juelich GmbH
!                        Markus Brenneis, Vladislav Kotov 
!
!    This file is part of EIRAM.
!
!    EIRAM is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    EIRAM is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with EIRAM.  If not, see <http://www.gnu.org/licenses/>.
!

!>
!> Wrappers are required for subroutines defined in a fortran module. 
!> See  eiram_octave.hpp, eiram_octave.cpp
!> as an example of how to call those subroutines in a C program
!>
!> @note Some functioins are turned into subroutines becazse because in Octave 
!>       the use of functions requires some extra work.
!> Quotation from the Octave documentation:
!>   Note that we don't attempt to handle Fortran functions, we always use
!>   subroutine wrappers for them and pass the return value as an extra
!>   argument.
!> The problem is that the results are written to a double* resultVar.fortran_vec(),
!> but a function would return a double on the stack, so the return value gets
!> invalidated when leaving the function.
!> On could work around this problem by explicitly copying the result to the octave variable,
!> but this is more complicated than using subroutines.

!> module-less wrapper for eiram_load
subroutine eiram_load_wrapper(filePath, fileName, err)
    use eiram
    implicit none
    character(*), intent(in) :: filePath, fileName
    integer, intent(out) :: err
    call eiram_load(filePath, fileName, err)
end subroutine

!> module-less wrapper for eiram_deallocate
subroutine eiram_deallocate_wrapper(err)
    use eiram
    implicit none
    integer, intent(out) :: err
    call eiram_deallocate(err)
end subroutine

!> module-less subroutine wrapper for eiram_getId
subroutine eiram_get_id_wrapper(fileName, reactionType, reactionIndex, id, err)
    use eiram
    implicit none
    character(*), intent(in) :: fileName, reactionType, reactionIndex
    integer, intent(out) :: id, err
    id = eiram_get_id(fileName, reactionType, reactionIndex, err)
end subroutine

!> subroutine wrapper for eiram_get_order
subroutine eiram_get_order_wrapper(N,M,Id,err)
    use eiram
    implicit none
    integer,intent(out) :: N,M
    integer,intent(in) :: Id 
    integer,intent(out) :: err
    call eiram_get_order(N,M,Id,err)
end subroutine eiram_get_order_wrapper

!no wrapper for eiram_create_imag because it uses data structure
!defined in a module

!function eiram_fit is defined outside a module - should not need a wrapper  

!> module-less wrapper for eiram_calc1_fast
function eiram_calc1_fast_wrapper(Id,lnX)
    use eiram
    implicit none
    integer, intent(in) :: Id
    real(eiram_dp), intent(in) :: lnX
    real(eiram_dp) :: eiram_calc1_fast_wrapper
    eiram_calc1_fast_wrapper=eiram_calc1_fast(Id,lnX)
end function eiram_calc1_fast_wrapper

!> module-less wrapper for eiram_calc1 
subroutine eiram_calc1_wrapper(Y,Id,lnX,N,err)
  use eiram
  implicit none
  real(eiram_dp), intent(out) :: Y(N) 
  integer, intent(in) :: Id
  real(eiram_dp), intent(in) :: lnX(N)
  integer, intent(out) :: err
  integer,intent(in) :: N
  call eiram_calc1(Y,Id,lnX,err)
end subroutine eiram_calc1_wrapper

!> module-less wrapper for eiram_calc12 
subroutine eiram_calc12_wrapper(B,Id,lnX,M,N,err)
  use eiram
  implicit none
  real(eiram_dp), intent(out) :: B(M,N) 
  integer, intent(in) :: Id
  real(eiram_dp), intent(in) :: lnX(N)
  integer, intent(out) :: err
  integer,intent(in) :: M,N 
  call eiram_calc12(B,Id,lnX,err)
end subroutine eiram_calc12_wrapper

!> module-less wrapper for eiram_calc2 
subroutine eiram_calc2_wrapper(Y,Id,LnX1,LnX2,N,err)
  use eiram
  implicit none
  real(eiram_dp), intent(out) :: Y(N) 
  integer, intent(in) :: Id
  real(eiram_dp), intent(in) :: lnX1(N)
  real(eiram_dp), intent(in) :: lnX2(N)
  integer, intent(out) :: err
  integer,intent(in) :: N  
  call eiram_calc2(Y,Id,LnX1,LnX2,err)
end subroutine eiram_calc2_wrapper

!> Binding for \ref eiram_calc1 which has to be called 
!> from Matlab/Octave interface
subroutine eiram_matlab_calc1(Y, X, N, fileName, reactionType, reactionIndex, err)
    use eiram
    implicit none
    real(eiram_dp),intent(out) :: Y(N) 
    real(eiram_dp), intent(in) :: X(N)
    integer, intent(in) :: N
    character(*), intent(in) :: fileName, reactionType, reactionIndex
    integer, intent(out) :: err
    integer :: Id
    intrinsic all,log 
    Id = eiram_get_id(fileName, reactionType, reactionIndex, err)
    if(err /= 0) return 
    if(.not. all(X>0.) ) then
     err=11
     return 
    end if
    call eiram_calc1(Y,Id,log(X),err)
end subroutine eiram_matlab_calc1

!> Binding for \ref eiram_calc2 which has to be called 
!> from Matlab/Octave interface
!>
!> WARNING: for reaction type H.4, H.10, H.12 first parameter -
!> particle deinsity - is automatically divided by 1e8
subroutine eiram_matlab_calc2(Y, X1, X2, M, N, fileName, reactionType, reactionIndex, err)
    use eiram
    implicit none
    real(eiram_dp),intent(out) :: Y(N,M)
    real(eiram_dp),intent(in) :: X1(M)
    real(eiram_dp),intent(in) :: X2(N)
    integer,intent(in) :: M,N
    character(*),intent(in) :: fileName, reactionType, reactionIndex
    integer,intent(out) :: err
    integer :: Id,j
    logical :: ldens
    intrinsic all,log 
    Id = eiram_get_id(fileName, reactionType, reactionIndex, err)
    if(err /= 0) return 
    if(.not. all(X1>0.) ) then
     err=12
     return 
    end if
    if(.not. all(X2>0.) ) then
     err=13
     return 
    end if

    if((reactionType.eq.'H.4').or.(reactionType.eq.'H.10').or.(reactionType.eq.'H.12')) then
     ldens=.true. !1st parameter is particle density
    else
     ldens=.false. 
    end if  

     if(ldens) then  
      do j=1,M
       call eiram_calc2(Y(:,j),Id,(/log(X1(j)/1.0e8)/),log(X2),err)       
      end do
     else
      do j=1,M
       call eiram_calc2(Y(:,j),Id,(/log(X1(j))/),log(X2),err)       
      end do
     end if

end subroutine eiram_matlab_calc2
