!> @file eiram_test.f90
!> Unit tests for eiram.f90
!> @authors Markus Brenneis Markus.Brenneis@uni-duesseldorf.de,
!>          Vladislav Kotov v.kotov@fz-juelich.de (contact person)
!>

!    Copyright (c) 2016  Forschungszentrum Juelich GmbH
!                        Markus Brenneis, Vladislav Kotov 
!
!    This file is part of EIRAM.
!
!    EIRAM is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    EIRAM is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with EIRAM.  If not, see <http://www.gnu.org/licenses/>.
!

module eiram_test
    use eiram
    implicit none

    public eiram_test_wrong_file
    public eiram_test_used_uninitialized
    public eiram_test_load 
    public eiram_test_wrong_ids
    public eiram_test_unknown_reaction
    public eiram_test_wrong_type
    public eiram_test_wrong_size
    public eiram_test_calculation
    public eiram_test_calc12
    public eiram_test_image
    intrinsic log,exp,tiny 
    
    private
    !> Maximum relative error, parameter which is used to compare two values 
    real(EIRAM_DP), parameter :: REL_TOLERANCE = 1e-7
    character(*), parameter :: DATA_FILE_PATH = "../data/"
    integer, parameter :: N_REACTIONS = 5
    integer, parameter :: N_ENERGY = 5
    character(9),parameter :: FILES(N_REACTIONS) =   (/ "HYDHELvk " , "AMJUELvk " , "HYDHELvk "  , "AMJUELvk ", "METHANEvk" /)
    character(6),parameter :: TYPES(N_REACTIONS) =   (/ "H.1"    , "H.2"    , "H.3"     , "H.4",   "H.2"    /)
    character(8),parameter :: INDICES(N_REACTIONS) = (/ "3.1.8  ", "2.4B0  ", "3.1.8  " , "2.1.5  ", "2.23   "/)
    real(EIRAM_DP),parameter :: ENERGY(N_ENERGY) =  (/ 5e-1, 1e0, 1e1, 1e2, 1e3 /)
    real(EIRAM_DP), parameter :: X1(N_REACTIONS) = (/ -1e0, -1e0, 1e1, 1e6, -1e0 /)
    !> Reference values for regression tests in eiram_test_calculation
    real(EIRAM_DP),parameter :: REFERENCE_VALUES(N_ENERGY,N_REACTIONS)= &
    reshape( (/ &
    ! HYDHEL H.1 3.1.8
    6.313479104896471d-15,  6.034774383818398d-15, &
    4.631473573509119d-15,  3.143444136405398d-15, 1.744077658482620d-15, &
    ! AMJUEL H.2 2.4B0
    1.229920711197272d-16, 1.857264485174120d-12, &
    2.944795004910767d-08, 1.073462933550498d-07, 7.992508213267738d-08, & 
    ! HYDHEL H.3 3.1.8
    2.064031139307574d-08, 2.084428058191167d-08, & 
    2.582052648196180d-08, 4.420072796238867d-08, 8.043311877821200d-08, &
    ! AMJUEL H.4 2.1.5
    1.865302180435326d-19, 4.904897322289924d-14, & 
    9.954471496437615d-09, 4.738209685016329d-08, 2.907993073331902d-08, &
    ! METHANE H.2 2.23
    1.320141447945264d-18, 9.051595057476462d-14, &
    2.966866133652682d-08, 1.909445571151299d-07, 1.391314913417069d-07 /), &
    (/N_ENERGY,N_REACTIONS/)) 
    contains
    
    !> Test for correct error codes if the input file does not exist or is malformed
    logical function eiram_test_wrong_file()
        integer :: err
        eiram_test_wrong_file = .false.
        call eiram_load(DATA_FILE_PATH, "AMJUELxx", err)
        if(err /= 310) goto 100
        call eiram_load("input/", "HYDHEL.corrupted", err)
        if(err /= 300) goto 100
        call eiram_deallocate(err)
        if(err /= 0) goto 100
        eiram_test_wrong_file = .true.
        return
100     print *, "ERROR in eiram_test_wrong_file: unexpected error code ", err
        call eiram_deallocate(err)
        return
    end function eiram_test_wrong_file
    
    !> Test for correct error codes if eiram is used without initialization
    logical function eiram_test_used_uninitialized()
        integer :: err,id,n,m
        real(eiram_dp) :: tmp(1),b(9,1)
        type(eiram_data),allocatable :: data(:)
        eiram_test_used_uninitialized = .false.
        id = eiram_get_id("HYDHEL", "H.1", "3.1.8", err)
        if(err /= 50) goto 150
        call eiram_get_order(n,m,1, err)
        if(err /= 50) goto 150
        call eiram_calc1(tmp,1,(/1._eiram_dp/),err)
        if(err /= 50) goto 150
        call eiram_calc12(b,1,(/2._eiram_dp/),err)
        if(err /= 50) goto 150
        call eiram_calc2(tmp,1,(/1._eiram_dp/),(/1._eiram_dp/),err)
        if(err /= 50) goto 150
        call eiram_create_image(data,err)
        if(err /= 50) goto 150
        call eiram_deallocate(err)
        if(err /= 0) goto 150
        eiram_test_used_uninitialized = .true.
        return
150     print *, "ERROR in eiram_test_used_uninitialized: unexpected error code ", err
        call eiram_deallocate(err)
        return
    end function eiram_test_used_uninitialized

    !> Technical test which shows that expected number of reactions is loaded from each input file
    logical function eiram_test_load()
     integer :: err,N,N0
     eiram_test_load=.false.
     call eiram_load(DATA_FILE_PATH, "HYDHEL", err)
     if(err /= 0) goto 100     
     N=eiram_nreact
     if(N/=208) then
      print *, "ERROR detected in eiram_test_load: "
      print *, " unexpected number of loaded reactions for HYDHEL"
      print *, " expected ",208," loaded ",N
      return  
     end if  
     N0=eiram_nreact
     call eiram_load(DATA_FILE_PATH, "HYDHELvk", err)
     if(err /= 0) goto 100
     N=eiram_nreact-N0
     if(N/=197) then
      print *, "ERROR detected in eiram_test_load: "
      print *, " unexpected number of loaded reactions for HYDHELvk"
      print *, " expected ",197," loaded ",N
      return  
     end if  
     N0=eiram_nreact
     call eiram_load(DATA_FILE_PATH, "AMJUELvk", err)
     if(err /= 0) goto 100
     N=eiram_nreact-N0
     if(N/=207) then
      print *, "ERROR detected in eiram_test_load: "
      print *, " unexpected number of loaded reactions for AMJUELvk"
      print *, " expected ",207," loaded ",N
      return  
     end if  
     N0=eiram_nreact
     call eiram_load(DATA_FILE_PATH, "AMJUEL", err)
     if(err /= 0) goto 100
     N=eiram_nreact-N0
     if(N/=273) then
      print *, "ERROR detected in eiram_test_load: "
      print *, " unexpected number of loaded reactions for AMJUEL"
      print *, " expected ",273," loaded ",N
      return  
     end if  
     N0=eiram_nreact
     call eiram_load(DATA_FILE_PATH, "METHANE", err)
     if(err /= 0) goto 100
     N=eiram_nreact-N0
     if(N/=91) then
      print *, "ERROR detected in eiram_test_load: "
      print *, " unexpected number of loaded reactions for METHANE"
      print *, " expected ",91," loaded ",N
      return  
     end if  
     N0=eiram_nreact
     call eiram_load(DATA_FILE_PATH, "H2VIBR", err)
     if(err /= 0) goto 100
     N=eiram_nreact-N0
     if(N/=112) then
      print *, "ERROR detected in eiram_test_load: "
      print *, " unexpected number of loaded reactions for H2VIBR"
      print *, " expected ",112," loaded ",N
      return  
     end if  
     N0=eiram_nreact
     call eiram_load(DATA_FILE_PATH, "METHANEvk", err)
     if(err /= 0) goto 100
     N=eiram_nreact-N0
     if(N/=64) then
      print *, "ERROR detected in eiram_test_load: "
      print *, " unexpected number of loaded reactions for METHANEvk"
      print *, " expected ",64," loaded ",N
      return  
     end if  
     call eiram_load(DATA_FILE_PATH, "H2VIBRvk", err)
     if(err /= 0) goto 100
     N=eiram_nreact-N0
     if(N/=161) then
      print *, "ERROR detected in eiram_test_load: "
      print *, " unexpected number of loaded reactions for H2VIBRvk"
      print *, " expected ",161," loaded ",N
      return  
     end if  

     call eiram_deallocate(err)
     if(err /= 0) goto 100

     eiram_test_load = .true.
     return

100  print *, "ERROR in eiram_test_load: unexpected error code ", err
     call eiram_deallocate(err)
     return
    end function eiram_test_load

    !> Test for correct error codes if reaction ID is invalid
    logical function eiram_test_wrong_ids()
        integer :: err
        real(kind=EIRAM_DP) :: tmp(1),b(9,1)        
        eiram_test_wrong_ids = .false.
        call eiram_load(DATA_FILE_PATH, "HYDHEL", err)
        if(err /= 0) goto 200
        call eiram_calc1(tmp,-1, (/1d0/),err)
        if(err /= 100) goto 200
        call eiram_calc1(tmp,100000,(/1d0/),err)
        if(err /= 100) goto 200
        call eiram_calc12(b,-1, (/2d0/),err)
        if(err /= 100) goto 200
        call eiram_calc12(b,100000,(/2d0/),err)
        if(err /= 100) goto 200
        call eiram_calc2(tmp,-1,(/1d0/),(/2d0/),err)
        if(err /= 100) goto 200
        call eiram_calc2(tmp,100000,(/1d0/),(/2d0/),err)
        if(err /= 100) goto 200
        call eiram_deallocate(err)
        if(err /= 0) goto 200
        eiram_test_wrong_ids = .true.
        return
200     print *, "ERROR in eiram_test_wrong_ids: unexpected error code ", err
        call eiram_deallocate(err)
        return
    end function eiram_test_wrong_ids    

    !> Test for correct error code when querying id for an unknown reaction
    logical function eiram_test_unknown_reaction()
        integer :: err, id
        eiram_test_unknown_reaction = .false.
        call eiram_load(DATA_FILE_PATH, "HYDHEL", err)
        if(err /= 0) goto 300
        id = eiram_get_id("HYDHEL", "H.1", "3.18", err)
        if(err /= 110) goto 300
        id = eiram_get_id("HYDHEL", "H1", "3.1.8", err)
        if(err /= 110) goto 300
        id = eiram_get_id("AMJUEL", "H.1", "3.1.8", err)
        if(err /= 110) goto 300
        id = eiram_get_id("AMMMJUEL", "H.1", "3.1.8", err)
        if(err /= 110) goto 300
        id = eiram_get_id("HYDHEL", "H.1", "3.1.8", err)
        if(err /= 0 .or. id < 1) goto 300
        call eiram_deallocate(err)
        if(err /= 0) goto 300
        eiram_test_unknown_reaction = .true.
        return
300     print *, "ERROR in eiram_test_unknown_reaction: unexpected error code ", err
        call eiram_deallocate(err)
        return
    end function eiram_test_unknown_reaction
    
    !> Test for correct error codes if _calc1 is called for double polynomial or
    !> _calc2, _calc12 is called for single polynomial  
    logical function eiram_test_wrong_type()
     integer :: err, id
     real(eiram_dp) :: tmp(1),b(9,1)
     eiram_test_wrong_type = .false.
     call eiram_load(DATA_FILE_PATH, "AMJUEL", err)
     if(err /= 0) goto 100
     id = eiram_get_id("AMJUEL","H.2","3.2.3",err)
     if(err /= 0) goto 100
     call eiram_calc2(tmp,id,(/1._eiram_dp/),(/2._eiram_dp/),err) 
     if(err /= 120) goto 100
     call eiram_calc12(b,id,(/2._eiram_dp/),err) 
     if(err /= 120) goto 100
     id = eiram_get_id("AMJUEL","H.4","2.1.5",err)
     if(err /= 0) goto 100
     call eiram_calc1(tmp,id,(/1._eiram_dp/),err) 
     if(err /= 130) goto 100
     call eiram_deallocate(err)
     if(err /= 0) goto 100
     eiram_test_wrong_type = .true.
     return
100  print *, "ERROR in eiram_test_wrong_type: unexpected error code ", err
     call eiram_deallocate(err)
     return
    end function eiram_test_wrong_type
     
    !> Test for correct error codes in case of mismatch of the array lengths
    logical function eiram_test_wrong_size()
     integer :: err, id
     real(eiram_dp) :: tmp(3),b(8,3),x1(2),x2(1),x(3),xx(1)
     eiram_test_wrong_size = .false.
     call eiram_load(DATA_FILE_PATH, "AMJUEL", err)
     if(err /= 0) goto 100
     call eiram_calc1(tmp,1,x1,err)
     if(err /= 100) goto 100
     call eiram_calc12(b,1,x1,err)
     if(err /= 100) goto 100
     id = eiram_get_id("AMJUEL","H.4","2.2.5g",err)
     call eiram_calc12(b,id,x,err)
     if(err /= 100) goto 100
     call eiram_calc2(tmp,1,x1,x2,err)
     if(err /= 100) goto 100
     call eiram_calc2(tmp,1,x,x2,err)
     if(err /= 100) goto 100
     call eiram_calc2(tmp,1,x1,x,err)
     if(err /= 100) goto 100
     call eiram_calc2(tmp,1,xx,x2,err)
     if(err /= 100) goto 100
     call eiram_deallocate(err)
     if(err /= 0) goto 100
     eiram_test_wrong_size = .true.
     return
100  print *, "ERROR in eiram_test_wrong_length: unexpected error code ", err
     call eiram_deallocate(err)
     return
    end function eiram_test_wrong_size
    
    !> Regression tests: compare calculated values with references
    logical function eiram_test_calculation()
        integer :: reactionId, r, err
        real(EIRAM_DP) :: calculatedValues(N_ENERGY)
        logical, dimension(N_ENERGY) :: passed

        intrinsic log

        eiram_test_calculation = .false.
        err = 0
        
        call eiram_load(DATA_FILE_PATH, "AMJUELvk", err)
        if(err /= 0) goto 1000
        call eiram_load(DATA_FILE_PATH, "HYDHELvk", err)
        if(err /= 0) goto 1000
        call eiram_load(DATA_FILE_PATH, "METHANEvk", err)
        if(err /= 0) goto 1000

        do r = 1, N_REACTIONS
            reactionId = eiram_get_id(FILES(r), TYPES(r), INDICES(r), err)
            if(err /= 0) goto 1000
            if(X1(r) < 0) then
             call eiram_calc1(calculatedValues,reactionId,log(ENERGY),err)
             if(err /= 0) goto 1000
            else            
             call eiram_calc2(calculatedValues,reactionId,(/log(X1(r))/),log(ENERGY),err)
             if(err /= 0) goto 1000
            end if  
            passed = isPassed(REFERENCE_VALUES(:,r), calculatedValues)
            call printFailed(REFERENCE_VALUES(:,r), calculatedValues, passed)
            if(.not.all(passed)) goto 1001
        end do

        eiram_test_calculation = .true.
        call eiram_deallocate(err)
        return

1000    print *, "ERROR in eiram_test_calculation: unexpected err code ", err
        call eiram_deallocate(err)
        return
1001    print *, "ERROR in eiram_test_calculation: calculated values do not match pre-calculated values"
        call eiram_deallocate(err)
        return
    end function eiram_test_calculation

    !> Regression test for subroutines _calc12 and _get_order
    !> compare _calc12 with subsequent call of _fit with _calc2  
    logical function eiram_test_calc12()
    integer :: N,M,err,ID,k
    integer,parameter :: NCELL=4
    real(eiram_dp) :: T(NCELL)=(/0.1,1.0,10.0,100.0/), &
                      E(NCELL)=(/1.0,1.0,10.0,10.0/)
    real(eiram_dp) :: b(9,NCELL),Y1(NCELL),Y2(NCELL)
    logical :: passed(NCELL)

    call eiram_load(DATA_FILE_PATH, "AMJUEL", err)
    if(err /= 0) goto 100
    eiram_test_calc12 = .false.
    ID=eiram_get_id("AMJUEL","H.3","0.3D",err) 
    if(err /= 0) goto 100
    call eiram_get_order(N,M,Id,err)
    if(err /= 0) goto 100
    if(N /= 8 .or. M /= 8) then
      print *, "ERROR: wrong values returned by eiram_get_order, N,M ",N,M
    end if 
    call eiram_calc2(Y1,ID,log(E),log(T),err)
    if(err /= 0) goto 100
    call eiram_calc12(B,ID,log(T),err)
    if(err /= 0) goto 100
    do k=1,NCELL
     Y2(k)=exp(eiram_fit(B(:,k),M,log(E(k))))
    end do

    passed = isPassed(Y1,Y2)
    call printFailed(Y1,Y2, passed)
    if(.not.all(passed)) goto 101

    eiram_test_calc12 = .true.
    call eiram_deallocate(err)
    return 
   
100 print *, "ERROR in eiram_test_calc12: unexpected err code ", err
    call eiram_deallocate(err)
    return
101 print *, "ERROR in eiram_test_calc12: mismatch between eiram_calc2 and eiram_calc12"
    call eiram_deallocate(err)
    return

    end function eiram_test_calc12

    !> Technical test of eiram_create_image
    logical function eiram_test_image()
    integer :: N,i,err
    type(eiram_data),allocatable :: data(:)

    eiram_test_image=.false.
    call eiram_load(DATA_FILE_PATH, "HYDHEL", err)
    if(err /= 0) goto 100
    call eiram_load(DATA_FILE_PATH, "AMJUEL", err)
    if(err /= 0) goto 100
    call eiram_load(DATA_FILE_PATH, "METHANEvk", err)
    if(err /= 0) goto 100

    call eiram_create_image(data,err)
    if(err /= 0) goto 100

    N=size(data)    

    do i=1,N
     deallocate(data(i)%creac,stat=err)
     if(err /= 0) goto 100
    end do  
    deallocate(data,stat=err)
    if(err /= 0) goto 100

    eiram_test_image = .true.
    call eiram_deallocate(err)
    return 

100 print *, "ERROR in eiram_test_image: unexpected err code ", err
    call eiram_deallocate(err)
    return
    end function eiram_test_image  


    !> @return true if \b expected and \b calculated 
    !> are same withing defined relative tolerance (REL_TOLERANCE)
    logical elemental function isPassed(expected, calculated)
        real(kind=EIRAM_DP), intent(in) :: expected, calculated
        intrinsic tiny
        real(kind=EIRAM_DP) :: eps
        eps=10.*tiny(expected)
        isPassed = abs(calculated - expected) < REL_TOLERANCE*expected + EPS
    end function isPassed
    
    !> Print pairs of values (expected(i),calculated(i)) 
    !> for which passed(i)=.false.
    !>
    !> This function is used to print the pairs which did not pass the check
    subroutine printFailed(expected, calculated, passed)
        real(kind=EIRAM_DP), dimension(:), intent(in) :: expected, calculated
        logical, dimension(size(expected)), intent(in) :: passed
        integer :: i
        do i = 1, size(expected)
            if(.not. passed(i)) then
                print *, "  Expected", expected(i), ", got", calculated(i)
                print *, "  Relative Error", abs(calculated(i) - expected(i))*100/expected(i), "%"
            end if
        end do
    end subroutine printFailed
    

end module eiram_test


program test
    use eiram_test
    use eiram
    implicit none

    eiram_unit = 0
    if(.not.eiram_test_wrong_file()) STOP "ERROR detected in eiram_test_wrong_file"
    if(.not.eiram_test_used_uninitialized()) STOP "ERROR detected in eiram_test_used_uninitialized"
    if(.not.eiram_test_load()) STOP "ERROR detected in eiram_test_load"
    if(.not.eiram_test_wrong_ids()) STOP "ERROR detected in eiram_test_wrong_ids"
    if(.not.eiram_test_unknown_reaction()) STOP "ERROR detected in eiram_test_unknown_reaction"
    if(.not.eiram_test_wrong_type()) STOP "ERROR detected in eiram_test_wrong_type"
    if(.not.eiram_test_calculation()) STOP "ERROR detected in eiram_test_calculation"
    if(.not.eiram_test_calc12()) STOP "ERROR detected in eiram_test_calc12"
    if(.not.eiram_test_image()) STOP "ERROR detected in eiram_test_image"
    print *, "EIRAM_TEST COMPLETED"
end program
