!> @file eiram.F
!> Matlab interface for EIRAM
!> @authors Markus Brenneis Markus.Brenneis@uni-duesseldorf.de,
!>          Vladislav Kotov v.kotov@fz-juelich.de (contact person)
!>

!    Copyright (c) 2016  Forschungszentrum Juelich GmbH
!                        Markus Brenneis, Vladislav Kotov 
!
!    This file is part of EIRAM.
!
!    EIRAM is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    EIRAM is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with EIRAM.  If not, see <http://www.gnu.org/licenses/>.
!

#include "fintrf.h"

!> Matlab interface for EIRAM
!>
!> @note Matlab keeps the memory *per mex function* unless <tt>clear myMexFun</tt>
!> or <tt>clear mex</tt> is called.
!>
!> @warning In Matlab, messages printed by EIRAM appear only on the console 
!> (matlab -nodesktop, not in the Matlab GUI window)  

      subroutine mexFunction(nlhs, plhs, nrhs, prhs)
      use eiram
      implicit none
      !> Number of vectors in the output (left hand side)
      integer :: nlhs
      !> Number of vectors in the input (right hand side)
      integer :: nrhs
      !> Output (left hand side) vectors
      mwpointer, dimension(*) :: plhs
      !> Input (right hand side) vectors
      !> 
      !> 1. Array of first arguments
      !> 2. (optional) Array of seconds arguments 
      !> 2(3). Name of the data-set file ('HYDHEL', 'AMJUEL', etc.)
      !> 3(4). Type of data ('H.1', etc.)
      !> 4(5). Array of strings with indices of reactions ('2.1.5', etc.)
      !> 5(6). Full path to the data (w/o file name) - optional
      mwpointer, dimension(*) :: prhs
      
      mwpointer :: mxGetPr, mxCreateDoubleMatrix, mxGetString
      integer :: mxIsNumeric, mxIsChar
      mwsize :: mxGetM, mxGetN
      
      mwsize :: nArg1s, nArg2s
      integer :: err,err0
      mwpointer :: Arg1s_pr, Arg2s_pr, Ys_pr
      real(kind=8),allocatable :: Arg1s(:), Arg2s(:)
      real(kind=8),allocatable :: Ys(:,:,:)
      integer, parameter :: MAX_STR_LEN = 1024
      character(MAX_STR_LEN) :: filePath,reactionIndexStr
      character(EIRAM_FILE_NAME_LENGTH) :: fileName
      character(EIRAM_TYPE_LENGTH) :: reactionType
      character(EIRAM_INDEX_LENGTH),allocatable :: reactionIndex(:)

      integer :: numberOfArgs,n_reactions,reactions_length,i,j,ind
      logical :: lPath,fileExists
      intrinsic trim,min

      if (nrhs < 4) 
     f    call mexErrMsgTxt('More than three inputs required.')


      if( mxIsNumeric(prhs(1)) /= 1 ) 
     f    call mexErrMsgTxt('Input # 1 must be a numeric array.')
      if( mxIsNumeric(prhs(2)) == 1 ) then 
        numberOfArgs = 2
      else 
        numberOfArgs = 1
      end if

      if( numberOfArgs == 1 ) then
       if (nrhs > 5) 
     f     call mexErrMsgTxt('Four or five inputs required.')
       if(mxIsChar(prhs(2)) /= 1) 
     f    call mexErrMsgTxt('Input # 2 must be a string.')
       if(mxIsChar(prhs(3)) /= 1) 
     f    call mexErrMsgTxt('Input # 3 must be a string.')
       if(mxIsChar(prhs(4)) /= 1) 
     f    call mexErrMsgTxt('Input # 4 must be a string.')
       if (nrhs > 4) then  
        if(mxIsChar(prhs(5)) /= 1) 
     f    call mexErrMsgTxt('Input # 5 must be a string.')      
         lPath=.true.
       else
         lPath=.false.  
       end if 
      end if 

      if( numberOfArgs == 2 ) then
       if (nrhs > 6 .or. nrhs < 5) 
     f     call mexErrMsgTxt('Five or six inputs required.')
       if(mxIsChar(prhs(3)) /= 1) 
     f    call mexErrMsgTxt('Input # 3 must be a string.')
       if(mxIsChar(prhs(4)) /= 1) 
     f    call mexErrMsgTxt('Input # 4 must be a string.')
       if(mxIsChar(prhs(5)) /= 1) 
     f    call mexErrMsgTxt('Input # 5 must be a string.')
       if (nrhs > 5) then  
        if(mxIsChar(prhs(6)) /= 1) 
     f    call mexErrMsgTxt('Input # 6 must be a string.')      
         lPath=.true.
       else
         lPath=.false.  
       end if 
      end if 
      
      nArg1s = mxGetN(prhs(1))
      if(mxGetM(prhs(1)) /= 1) then
          call mexErrMsgTxt('Input #1 must be a row vector.')
      endif

      Arg1s_pr = mxGetPr(prhs(1))
      if(numberOfArgs == 2) then
          Arg2s_pr = mxGetPr(prhs(2))
          nArg2s = mxGetN(prhs(2))
          if(mxGetM(prhs(2)) /= 1) then
              call mexErrMsgTxt('Input #2 must be a row vector.')
          endif
      else
          nArg2s = 0
      endif

      n_reactions = mxGetM(prhs(numberOfArgs+3))
      reactions_length = mxGetN(prhs(numberOfArgs+3))

      allocate(reactionIndex(n_reactions))

      err=mxGetString(prhs(numberOfArgs+1),fileName,
     ,                EIRAM_FILE_NAME_LENGTH)
      err=mxGetString(prhs(numberOfArgs+2),
     ,                     reactionType,EIRAM_TYPE_LENGTH)
      err=mxGetString(prhs(numberOfArgs+3),
     ,                     reactionIndexStr,MAX_STR_LEN)
      if(lPath) then 
       err=mxGetString(prhs(numberOfArgs+4),filePath, MAX_STR_LEN)
      else
       filePath=""
      end if
      !for reaction index divide long string into substrings
      reactionIndex = " "
      do i = 1, n_reactions
        do j = 1, min(reactions_length,EIRAM_INDEX_LENGTH)
           ind = (j-1)*n_reactions+i
           reactionIndex(i)(j:j) = reactionIndexStr(ind:ind)
        end do 
      end do

      !convert 1st (and 2nd - if required) argument into array of reals
      allocate(Arg1s(nArg1s))
      call mxCopyPtrToReal8(Arg1s_pr, Arg1s, nArg1s)
      if(nArg2s >0) then
         allocate(Arg2s(nArg2s))
         call mxCopyPtrToReal8(Arg2s_pr, Arg2s, nArg2s)
      end if
            
      eiram_unit = 0
      
      call eiram_load(filePath, fileName, err)
      if(err /= 0) then
        call eiram_deallocate(err0)
        select case(err)
         case(310) 
           call mexErrMsgTxt('Input file '//trim(filePath)
     /                      //trim(fileName)//' is not found.')
         case(300)
           call mexErrMsgTxt('Cannot read input file '
     /                     //trim(filePath)//trim(fileName))
         case default
          call mexErrMsgTxt("error during initialization of EIRAM")
        end select 
      end if
      
      if (nArg2s>0) then
       allocate(Ys(nArg2s,nArg1s,n_reactions))
      else
       allocate(Ys(nArg1s,1,n_reactions))
      end if

      do i=1,n_reactions      
       if(nArg2s > 0) then  
        call eiram_matlab_calc2(Ys(:,:,i),Arg1s,Arg2s,nArg1s,nArg2s,
     &                    fileName,reactionType,reactionIndex(i),err)
       else 
         call eiram_matlab_calc1(Ys(:,1,i),Arg1s,nArg1s,
     &               fileName,reactionType,reactionIndex(i),err)
       end if ! if(nArg2s > 0) then  
       if(err /= 0) call eiram_deallocate(err0)
       select case(err)
       case(11)
         call mexErrMsgTxt("error : negative argument")
       case(12)
         call mexErrMsgTxt("error : negative 1st argument")
       case(13)
         call mexErrMsgTxt("error : negative 2nd argument")
       case(110)
         call mexErrMsgTxt("error: requested reaction "
     /            //trim(fileName)//" "//trim(reactionType)//" "
     /           //trim(reactionIndex(i))//" is not found")
       case(120)
         call mexErrMsgTxt("error: double fit is requested for reaction"
     /                     //" described by a single fit")
       case(130)
         call mexErrMsgTxt("error: single fit is requested for reaction"
     /                    //" described by a double fit")
       case default
         if(err/=0) call mexErrMsgTxt("error in EIRAM")
       end select
      end do !do i=1,n_combinations

      do i=1,n_reactions
       if(nArg2s > 0) then  
        plhs(i) = mxCreateDoubleMatrix(nArg2s,nArg1s,0)
       else
        plhs(i) = mxCreateDoubleMatrix(nArg1s,1,0)
       end if
       Ys_pr = mxGetPr(plhs(i))
       call mxCopyReal8ToPtr(Ys(:,:,i), Ys_pr, max(1,nArg2s)*nArg1s)
      end do 

      call eiram_deallocate(err)

      deallocate(reactionIndex)      
      deallocate(Arg1s)
      if(allocated(Arg2s)) deallocate(Arg2s)
      deallocate(Ys)
      
      end subroutine mexFunction
