!> @file eiram_cgi.f90
!> CGI (Common Gateway Interface) for EIRAM
!> @authors Markus Brenneis Markus.Brenneis@uni-duesseldorf.de,
!>          Vladislav Kotov v.kotov@fz-juelich.de (contact person)
!>
!> CGI (Common Gateway Interface) for EIRAM
!>
!> GET request parameters
!>
!> files: comma separated list of file names
!>
!> types: comma separated list of type names
!>
!> reactions: comma separated list of reaction indices
!>
!> arg1s: comma separated list of first arguments, may be single values or vectors in start[:delta]:end notation
!>
!> arg2s: comma separated list of second arguments, may be "-", 
!>        single values or vectors in start[:delta]:end notation; 
!>        if arg2 is "-", tnen only one argument arg1 taken for this reaction
!>
!> epsfilter: real number; tolerance parameter for the energy filter.
!>            The filter takes away nodes with values which are equal within relative accuracy 
!>            defined by epsfilter. [default: epsfilter=0, ie. no filtering]
!>
!> All lists must have the same length.
!>
!> If the query string does not contain '=', 
!> then a JSON (JavaScript Object Notation) format list of available combinations is returned.
!>
!> In case of an error 400 Bad Request is returned. 
!> The body of the response contains a rough description of the error.
!>
!> examples for query strings:
!> * QUERY_STRING="files=AMJUEL&types=H.1&reactions=0.1D&arg1s=100&arg2s=-"
!> * QUERY_STRING="files=HYDHEL,AMJUEL&types=H.1,H.3&reactions=3.1.8,0.2T&arg1s=1e-2:1e-2:1e-1,1e1&arg2s=-,0.37e0"
!> * QUERY_STRING="files=AMJUEL,AMJUEL&types=H.3,H.4&reactions=0.2T,2.1.5&arg1s=1d3,1d10&arg2s=3.7d-1,1d0:10d0"
!> * QUERY_STRING="files=AMJUEL,AMJUEL&types=H.3,H.4&reactions=0.2T,2.1.5&arg1s=1d3,1d10&arg2s=3.7d-1,1d0:100d0&epsfilter=1e-2"
!>

!    Copyright (c) 2016  Forschungszentrum Juelich GmbH
!                        Markus Brenneis, Vladislav Kotov 
!
!    This file is part of EIRAM.
!
!    EIRAM is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    EIRAM is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with EIRAM.  If not, see <http://www.gnu.org/licenses/>.
!

program eiram_cgi

    use cgi_protocol
    use eiram
    implicit none
    !> unit into which the JSON output is written
    integer                    :: unit
    !> dictionary which contains values passed to the programm (via GET or POST)
    type(dict_struct), pointer :: dict => null()
    !> header containing the type of the output and a newline
    character(*), parameter    :: HEADER = "Content-type: application/json; charset=utf-8"//char(10)//&
                                    "Content-Disposition: attachment; filename=data.json"//char(10)
    
    logical, parameter :: DEBUG = .false.
    integer, parameter :: DEBUG_UNIT = 0
    
    !> maximum length of a get parameter
    !> Note that it must be big enough to hold a list of vectors like 63.95456031906131:0.6283673168432119:126.7912920033825.
    integer, parameter :: MAX_PARAM_LENGTH = 500
    integer, parameter :: TMP_UNIT = 142
    real(kind=EIRAM_DP) :: NEG_INF = -huge(1.d0)
    character(MAX_PARAM_LENGTH) :: input = ""
    character(MAX_PARAM_LENGTH), dimension(:), allocatable :: files, types, reactions, arg1s, arg2s
    integer :: nFiles, nTypes, nReactions, nArg1s, nArg2s
    integer :: i, err
    real(kind=EIRAM_DP), dimension(:), allocatable :: xs, ys
    real(kind=EIRAM_DP) :: epsfilter = 0d0
    !> maximum number of points which are allowed in one request
    integer, parameter :: MAX_POINTS = 1000000

    open(TMP_UNIT, iostat=err, status='scratch', action='readwrite')
    eiram_unit = TMP_UNIT
    write(TMP_UNIT, *) "BEGIN OF LOG"
    write(TMP_UNIT, *) "CGI: not initialized"
    
    call get_environment_variable("QUERY_STRING", input)
    if(index(input, '=') == 0) then
        call sendAvailableReactions()
    end if
    
    call cgi_begin(output_no_header, dict, unit)
    
    call getCsvParameter(dict, "files", "no file names provided", files, nFiles)
    call getCsvParameter(dict, "types", "no type names provided", types, nTypes)
    call getCsvParameter(dict, "reactions", "no reaction indices provided", reactions, nReactions)
    call getCsvParameter(dict, "arg1s", "no first arguments provided", arg1s, nArg1s)
    call getCsvParameter(dict, "arg2s", "no second arguments provided", arg2s, nArg2s)
    call cgi_get(dict, "epsfilter", epsfilter)
    
    if(DEBUG) write(DEBUG_UNIT, *) nFiles, files
    if(DEBUG) write(DEBUG_UNIT, *) nTypes, types
    if(DEBUG) write(DEBUG_UNIT, *) nReactions, reactions
    if(DEBUG) write(DEBUG_UNIT, *) nArg1s, arg1s
    if(DEBUG) write(DEBUG_UNIT, *) nArg2s, arg2s
    
    if(.not.all((/nFiles, nTypes, nReactions, nArg1s/) == nArg2s)) then
        call cgi_error("esputr_cgi: all lists must have the same length")
    end if
    
    write(unit, '(A)') HEADER
    
    call initAllFiles()
    write(unit, '(A)'), "["
    do i = 1, nFiles
        if(i > 1) then
            write(unit, '(A)'), ","
        end if
        call getPoints(files(i), types(i), reactions(i), arg1s(i), arg2s(i), xs, ys)
        write(unit, '(A)') trim(pointsToJsonString(files(i), types(i), reactions(i), xs, ys))
        deallocate(xs, ys)
    end do
    write(unit, '(A)'), "]"
    
    call eiram_deallocate(err)
    call cgi_end()
    
    contains
    
    !> Calculate the points for the given reaction and arguments
    subroutine getPoints(fileName, reactionType, reactionIndex, arg1s_str, arg2s_str, xs, ys)
        !> file name for the reaction
        character(*), intent(in) :: fileName
        !> type name for the reaction
        character(*), intent(in) :: reactionType
        !> index of the reaction
        character(*), intent(in) :: reactionIndex
        !> argument 1, single numeric value or min:delta:max string
        character(*), intent(in) :: arg1s_str
        !> argument 2, "-", single numeric value or min:delta:max string
        !> if arg2 /= "-", then only the first value from arg1s is used
        character(*), intent(in) :: arg2s_str
        !> the x values for which the function is calculated
        real(kind=EIRAM_DP), allocatable, intent(out) :: xs(:)
        !> the y values
        real(kind=EIRAM_DP), allocatable, intent(out) :: ys(:)
        real(kind=EIRAM_DP), allocatable :: ys0(:)
        real(kind=EIRAM_DP), allocatable :: arg1s(:), arg2s(:)
        logical,allocatable :: filterMask(:)
        integer :: err,n,i,lastITaken

        intrinsic pack,size

!>
!> \todo the use of memory here could be further optimized
!>

        err = 0
        
        call parseVector(arg1s_str, arg1s)
        call parseVector(arg2s_str, arg2s)
        if(size(arg2s) == 0 .or. size(arg1s) == 0) call cgi_error("esputr_cgi: invalid input vectors")

        if(arg2s(1) <= NEG_INF) then
            allocate(ys0(size(arg1s)), stat=err)
            if(err /= 0) call cgi_error("esputr_cgi: memory allocation error"//getLastMessage())
            call eiram_matlab_calc1(ys0,arg1s, size(arg1s), fileName, reactionType, reactionIndex, err)
           if(err /= 0) call cgi_error("esputr_cgi: error while calculating values"//getLastMessage())
        else
            allocate(ys0(size(arg2s)), stat=err)
            if(err /= 0) call cgi_error("esputr_cgi: memory allocation error"//getLastMessage())
            call eiram_matlab_calc2(ys0, arg1s(1), arg2s, 1, size(arg2s), fileName, reactionType, reactionIndex, err)
            if(err /= 0) call cgi_error("esputr_cgi: error while calculating values"//getLastMessage())
        end if

        n=size(ys0)
        allocate(filterMask(n),stat=err)
        if(err /= 0) call cgi_error("esputr_cgi: memory allocation error"//getLastMessage())     

        filterMask(1) = .true.
        lastITaken = 1
        do i = 2, n-1
            filterMask(i) = abs(ys0(i)-ys0(lastITaken)) > .5*epsfilter*(abs(ys0(i))+abs(ys0(lastITaken)))
            if(filterMask(i)) lastITaken = i
        end do
        filterMask(n) = .true.

        n=count(filterMask)
        if(allocated(xs)) deallocate(xs)   
        if(allocated(ys)) deallocate(ys)   
        allocate(xs(n),ys(n),stat=err)
        if(err /= 0) call cgi_error("esputr_cgi: memory allocation error"//getLastMessage())
        if(arg2s(1) <= NEG_INF) then
         xs=pack(arg1s,filterMask)
        else 
         xs=pack(arg2s,filterMask)
        end if
        ys=pack(ys0,filterMask)

        deallocate(arg1s,arg2s,filterMask,ys0)

    end subroutine getPoints
    
    !> Return a string which represents a JSON object suitable for plotting of the given reaction data
    function pointsToJsonString(fileName, reactionType, reactionIndex, xs, ys)
        use json
        !> file name for the reaction
        character(*), intent(in) :: fileName
        !> type name for the reaction
        character(*), intent(in) :: reactionType
        !> index of the reaction
        character(*), intent(in) :: reactionIndex
        !> x values
        real(kind=EIRAM_DP), dimension(:), intent(in) :: xs
        !> y values
        real(kind=EIRAM_DP), dimension(:), intent(in) :: ys
        character(150+2*(14*size(xs)+2)) :: pointsToJsonString
        integer :: c, l
        pointsToJsonString(1:1) = "{"
        c = 2
       
        pointsToJsonString(c:c+3) = '"x":'
        c = c+4
        write(pointsToJsonString(c:), '(A)') eiram_json_getArrayAsJson(xs, l, 5)
        c = c+l
        
        pointsToJsonString(c:c+4) = ',"y":'
        c = c+5
        write(pointsToJsonString(c:), '(A)') eiram_json_getArrayAsJson(ys, l, 5)
        c = c+l
        
        call eiram_json_writeJsonStringProperty("file", fileName, pointsToJsonString, c)
        call eiram_json_writeJsonStringProperty("type", reactionType, pointsToJsonString, c)
        call eiram_json_writeJsonStringProperty("index", reactionIndex, pointsToJsonString, c)
        
        pointsToJsonString(c:c) = "}"
        c = c+1
    end function pointsToJsonString
    
    !> Parse values of a comma separated CGI parameter.
    subroutine getCsvParameter(dict, parameterName, errMsg, values, nValues)
        !> the CGI dictionary
        type(dict_struct), pointer, intent(in) :: dict
        !> the name of the parameter
        character(*), intent(in) :: parameterName
        !> the cgi error message returned in case the parameter does not exist
        character(*), intent(in) :: errMsg
        !> the array which will hold the parsed values
        character(MAX_PARAM_LENGTH), dimension(:), allocatable, intent(inout) :: values
        !> number of obtained values
        integer, intent(out) :: nValues
        character(DICT_VALUE_LENGTH) :: input = ""
        call cgi_get(dict, parameterName, input)
        if(input == "") call cgi_error("esputr_cgi: "//errMsg)
        call split(input, ",", values)
        nValues = size(values)
    end subroutine
    
    !> Send the list of all available reactions to the server
    subroutine sendAvailableReactions()
        integer :: err,i,N
        type(eiram_data),allocatable :: REACTIONS(:)

        unit = 6
        
        call initAllFiles()

        err=0 
        call eiram_create_image(REACTIONS,err)
        if(err /= 0) then
         call cgi_error("esputr_cgi: cannot create the list of reactions")
         return
        end if 

        N=size(REACTIONS)
        write(unit, '(A)') HEADER
        write(unit, '(A)') trim(getReactionsAsJson(REACTIONS,N,err))
        
        if(err /= 0) call cgi_error("eiram_cgi: error while getting combinations "//getLastMessage())
        
        do i=1,N
         deallocate(REACTIONS(i)%creac)
        end do  
        deallocate(REACTIONS)

        call eiram_deallocate(err)
        call cgi_end()
    end subroutine sendAvailableReactions

    !> Initialize EIRAM with all available data
    subroutine initAllFiles()
        call eiram_load("../data/", "H2VIBR", err)
        if(err /= 0) call cgi_error("eiram_cgi: initialization from H2VIBR failed"//getLastMessage())
        call eiram_load("../data/", "METHANE", err)
        if(err /= 0) call cgi_error("eiram_cgi: initialization from METHANE failed"//getLastMessage())
        call eiram_load("../data/", "AMJUEL", err)
        if(err /= 0) call cgi_error("eiram_cgi: initialization from AMJUEL failed"//getLastMessage())
        call eiram_load("../data/", "HYDHEL", err)
        if(err /= 0) call cgi_error("eiram_cgi: initialization from HYDHEL failed"//getLastMessage())
    end subroutine initAllFiles
    
    !> Return two last lines printed into the temporary output unit    
    character(DICT_VALUE_LENGTH*2) function getLastMessage()
        getLastMessage = ""
        backspace(TMP_UNIT, iostat=err)
        backspace(TMP_UNIT, iostat=err)
        read(TMP_UNIT, '(A)') getLastMessage(1:DICT_VALUE_LENGTH)
        read(TMP_UNIT, '(A)') getLastMessage(len_trim(getLastMessage)+1:DICT_VALUE_LENGTH*2)
    end function
    
    !> Split one line into individual strings for each comma separated substring
    subroutine split(string, c, array)
        !> c-separated string; must not contain commas (unless c is a comma) or whitespaces
        character(*), intent(in) :: string
        !> character at which the input string is split
        character(1), intent(in) :: c
        character(MAX_PARAM_LENGTH), dimension(:), allocatable, intent(out) :: array
        integer :: nSeps, i
        character(DICT_VALUE_LENGTH) :: str
        str = string
        nSeps = count(transfer(str, 'a', DICT_VALUE_LENGTH) == c)
        allocate(array(nSeps + 1))
        if(c /= ',') then
            do i=1, len(str)
                if(str(i:i) == c) then
                    str(i:i) = ","
                end if
            end do
        end if
        read(str, *) array(1:nSeps+1)
    end subroutine split
    
    !> Converts octave-like string vector representation into an array of reals
    subroutine parseVector(strArr, realArr)
        !> string in the form "<real>", "<real>:<real>", "<real>:<real>:<real>", or "-"
        !> "-" will be converted to NEG_INF
        character(*), intent(in) :: strArr
        real(kind=EIRAM_DP), dimension(:), allocatable, intent(out) :: realArr
        character(MAX_PARAM_LENGTH), dimension(:), allocatable :: strArrSplit
        real(kind=EIRAM_DP) :: start, delta, ende
        integer :: i, err, nValues
        
        if(allocated(realArr)) deallocate(realArr)

        if(strArr == "-") then
            allocate(realArr(1))
            realArr(1) = NEG_INF
            return
        end if
          
        if(allocated(strArrSplit)) deallocate(strArrSplit)    
        call split(strArr, ":", strArrSplit)
        select case(size(strArrSplit))
            case (1)
                allocate(realArr(1))
                read(strArrSplit(1), *, iostat=err) realArr(1)
                if(err/=0) call cgi_error("esputr_cgi: invalid real value "//strArrSplit(1))
                return
            case (2)
                read(strArrSplit(1), *, iostat=err) start
                if(err/=0) call cgi_error("esputr_cgi: invalid real value "//strArrSplit(1))
                read(strArrSplit(2), *, iostat=err) ende
                if(err/=0) call cgi_error("esputr_cgi: invalid real value "//strArrSplit(2))
                delta = 1d0
            case (3)
                read(strArrSplit(1), *, iostat=err) start
                if(err/=0) call cgi_error("esputr_cgi: invalid real value "//strArrSplit(1))
                read(strArrSplit(2), *, iostat=err) delta
                if(err/=0) call cgi_error("esputr_cgi: invalid real value "//strArrSplit(2))
                read(strArrSplit(3), *, iostat=err) ende
                if(err/=0) call cgi_error("esputr_cgi: invalid real value "//strArrSplit(3))
            case default
                call cgi_error("esputr_cgi: invalid vector format "//strArr)
        end select
        nValues = floor((ende-start)/delta) + 1
        if(nValues>MAX_POINTS) then 
         call cgi_error("eiram_cgi: too many data points are requested")
        end if 
        allocate(realArr(nValues))
        realArr = (/ ( start + i*delta, i = 0, nValues-1 ) /)
    end subroutine
    
   !> Transfer reactions stored in  the arrays REACTIONS
   !> into a string which represents a JSON (JavaScript Object Notation) object - list of reactions
    function getReactionsAsJson(REACTIONS,N,err)
        use json
        character(1024*N) :: getReactionsAsJson
        type(eiram_data),intent(in) :: REACTIONS(N)
        integer,intent(in) :: N 
        character(EIRAM_FILE_NAME_LENGTH) :: lastFileName
        character(EIRAM_TYPE_LENGTH) :: lastTypeName
        integer, intent(out) :: err
        integer :: r, c, l, m

        err=0

        lastFileName = ""
        lastTypeName = ""
        getReactionsAsJson(1:10) = '{"files":['
        c = 11
        forEachReaction: do r = 1, size(REACTIONS)
            
            if(lastFileName /= REACTIONS(r)%FILE) then
                if(lastFileName /= "") then
                    getReactionsAsJson(c:c+4) = "]}]},"
                    c = c+5
                end if
                lastFileName = REACTIONS(r)%FILE
                lastTypeName = ""
                getReactionsAsJson(c:c+8) = '{"name":"'
                c = c+9
                getReactionsAsJson(c:) = trim(REACTIONS(r)%FILE)
                c = c+len_trim(REACTIONS(r)%FILE)
                getReactionsAsJson(c:c+10) = '","types":['
                c = c+11
            end if
            if(lastTypeName /= REACTIONS(r)%TYPE) then
                if(lastTypeName /= "") then
                    getReactionsAsJson(c:c+2) = "]},"
                    c = c+3
                end if
                lastTypeName = REACTIONS(r)%TYPE
                getReactionsAsJson(c:c+8) = '{"name":"'
                c = c+9
                getReactionsAsJson(c:) = trim(REACTIONS(r)%TYPE)
                c = c+len_trim(REACTIONS(r)%TYPE)
                getReactionsAsJson(c:c) = '"'
                c = c+1
                
                call eiram_json_writeJsonStringProperty("yUnit", REACTIONS(r)%YUNITS, getReactionsAsJson, c)
                call eiram_json_writeJsonStringProperty("yQuantity", REACTIONS(r)%YQUANTITY, getReactionsAsJson, c)
                
                getReactionsAsJson(c:c+11) = ',"xUnits":["'
                c = c+12
                getReactionsAsJson(c:) = REACTIONS(r)%X1UNITS
                c = c+len_trim(REACTIONS(r)%X1UNITS)
                if(REACTIONS(r)%M > 0) then
                    getReactionsAsJson(c:c+2) = '","'
                    c = c+3
                    getReactionsAsJson(c:) = REACTIONS(r)%X2UNITS
                    c = c+len_trim(REACTIONS(r)%X2UNITS)
                end if
                getReactionsAsJson(c:c+1) = '"]'
                c = c+2
                
                getReactionsAsJson(c:c+16) = ',"xQuantities":["'
                c = c+17
                getReactionsAsJson(c:) = REACTIONS(r)%X1QUANTITY
                c = c+len_trim(REACTIONS(r)%X1QUANTITY)
                if(REACTIONS(r)%M > 0) then
                    getReactionsAsJson(c:c+2) = '","'
                    c = c+3
                    getReactionsAsJson(c:) = REACTIONS(r)%X2QUANTITY
                    c = c+len_trim(REACTIONS(r)%X2QUANTITY)
                end if
                getReactionsAsJson(c:c+1) = '"]'
                c = c+2
                
                getReactionsAsJson(c:c+4) = ',"n":'
                c = c+5
                write(getReactionsAsJson(c:c+1), '(I2)') REACTIONS(r)%N
                c = c+2
                
                getReactionsAsJson(c:c+4) = ',"m":'
                c = c+5
                write(getReactionsAsJson(c:c+2), '(I2)') REACTIONS(r)%M
                c = c+2
                
                getReactionsAsJson(c:c+13) = ',"reactions":['
                c = c+14
            else if(r > 1) then
                getReactionsAsJson(c:c) = ","
                c = c+1
            end if
            getReactionsAsJson(c:c) = "{"
            c = c+1
            
            write(getReactionsAsJson(c:c+7), '(A)') '"creac":'
            c = c+8
            if(allocated(REACTIONS(r)%CREAC)) then
                if(REACTIONS(r)%M == 0) then
                    write(getReactionsAsJson(c:), '(A)') eiram_json_getArrayAsJson(REACTIONS(r)%CREAC(:,0), l, 12)
                    c = c+l
                else
                    write(getReactionsAsJson(c:), '(A)') '['
                    c = c+1
                    do m = 0, REACTIONS(r)%M
                        if(m > 0) then
                            getReactionsAsJson(c:c) = ","
                            c = c+1
                        end if
                        write(getReactionsAsJson(c:), '(A)') eiram_json_getArrayAsJson(REACTIONS(r)%CREAC(:,m), l, 12)
                        c = c+l
                    end do
                    write(getReactionsAsJson(c:), '(A)') ']'
                    c = c+1
                end if
            else
                write(getReactionsAsJson(c:c+3), '(A)') "null"
                c = c+4
            end if
            
            call eiram_json_writeJsonStringProperty("equation", REACTIONS(r)%REACTION, getReactionsAsJson, c)
            call eiram_json_writeJsonStringProperty("index", REACTIONS(r)%INDEX, getReactionsAsJson, c)
            
            getReactionsAsJson(c:c+1) = "}"//char(10)
            c = c+2
        
        end do forEachReaction
        
        getReactionsAsJson(c:c+5) = "]}]}]}"
        c = c + 6

        if(DEBUG) print *, c, "/", len(getReactionsAsJson), "chars written"
    end function getReactionsAsJson

end program eiram_cgi
